# -*- coding: utf-8 -*-
"""This module defines options common to anm/gnm/pca commands."""

DEFAULTS = {}
HELPTEXT = {}

FIGFORMATS = 'eps pdf png ps raw rgba svg svgz'.split()

for key, txt, val in [
    ('outdir', 'output directory', '.'),
    ('prefix', 'output file prefix', ''),
    ('select', 'atom selection',
         "protein and name CA or nucleic and name P C4' C2"),
    ('nmodes', 'number of non-zero eigenvectors (modes) to calculate', 10),
    ('membrane', 'whether to include the explicit membrane model', False),

    ('outall', 'write all outputs', False),
    ('outeig', 'write eigenvalues/vectors', False),
    ('outcov', 'write covariance matrix', False),
    ('outnpz', 'write compressed ProDy data file', False),
    ('npzmatrices', 'write matrix to compressed ProDy data file', False),
    ('outscipion', 'write continuousflex modes directory and sqlite', False),
    ('outcc', 'write cross-correlations', False),
    ('outhm', 'write cross-correlations heatmap file', False),
    ('outsf', 'write square-fluctuations', False),
    ('extend', 'write NMD file for the model extended to '
               '"backbone" ("bb") or "all" atoms of the residue,'
               ' model must have one node per residue', ''),

    ('numformat', 'number output format', '%12g'),
    ('numext', 'numeric file extension', '.txt'),
    ('numdelim', 'number delimiter', ' '),


    ('figformat', 'figure file format', 'pdf'),
    ('figwidth', 'figure width (inch)', 8.),
    ('figheight', 'figure height (inch)', 6.),
    ('figdpi', 'figure resolution (dpi)', 300),
    ('figall', 'save all figures', False),
    ('figcc', 'save cross-correlations figure', False),
    ('figsf', 'save square-fluctuations figure', False),
    ('nproc', 'number of processors', 0),]:

    DEFAULTS[key] = val
    HELPTEXT[key] = txt


__all__ = ['addNMAParameters', 'addNMAOutput', 'addNMAOutputOptions',
           'addNMAFigures', 'addNMAFigureOptions']


def addNMAParameters(parser, include_nproc=True):

    parser = parser.add_argument_group('parameters')

    parser.add_argument('-n', '--number-of-modes', dest='nmodes', type=int,
        default=DEFAULTS['nmodes'], metavar='INT',
        help=HELPTEXT['nmodes'] + ' (default: %(default)s)')

    parser.add_argument('-s', '--select', dest='select', type=str,
        default=DEFAULTS['select'], metavar='STR',
        help=HELPTEXT['select'] + ' (default: "%(default)s")')
    
    parser.add_argument('-N', '--membrane', dest='membrane', action='store_true',
        default=DEFAULTS['membrane'], help=HELPTEXT['membrane'])
    
    if include_nproc:
        parser.add_argument('-P', '--number-of-processors', dest='nproc', type=int,
            default=DEFAULTS['nproc'], metavar='INT',
            help=HELPTEXT['nproc'] + ' (default: %(default)s meaning use all)')

    return parser


def addNMAOutput(parser):

    parser = parser.add_argument_group('output')

    parser.add_argument('-a', '--all-output',
        dest='outall', action='store_true',
        default=DEFAULTS['outall'], help=HELPTEXT['outall'])

    parser.add_argument('-o', '--output-dir', dest='outdir', type=str,
        default=DEFAULTS['outdir'], metavar='PATH',
        help=HELPTEXT['outdir'] + ' (default: %(default)s)')

    parser.add_argument('-e', '--eigenvs', dest='outeig', action='store_true',
        default=DEFAULTS['outeig'], help=HELPTEXT['outeig'])

    parser.add_argument('-r', '--cross-correlations', dest='outcc',
        action='store_true',
        default=DEFAULTS['outcc'], help=HELPTEXT['outcc'])

    parser.add_argument('-u', '--heatmap', dest='outhm',
        action='store_true',
        default=DEFAULTS['outhm'], help=HELPTEXT['outhm'])

    parser.add_argument('-q', '--square-fluctuations', dest='outsf',
        action='store_true',
        default=DEFAULTS['outsf'], help=HELPTEXT['outsf'])

    parser.add_argument('-v', '--covariance', dest='outcov',
        action='store_true',
        default=DEFAULTS['outcov'], help=HELPTEXT['outcov'])

    parser.add_argument('-z', '--npz', dest='outnpz', action='store_true',
        default=DEFAULTS['outnpz'], help=HELPTEXT['outnpz'])

    parser.add_argument('-Z', '--npzmatrices', dest='npzmatrices', action='store_true',
        default=DEFAULTS['npzmatrices'], help=HELPTEXT['npzmatrices'])

    parser.add_argument('-S', '--export-scipion', dest='outscipion', action='store_true',
        default=DEFAULTS['outscipion'], help=HELPTEXT['outscipion'])

    parser.add_argument('-t', '--extend', dest='extend', type=str,
        metavar='STR', choices=set(['bb', 'all', 'backbone']),
        help=HELPTEXT['extend'])

    return parser


def addNMAOutputOptions(parser, prefix):

    parser = parser.add_argument_group('output options')

    parser.add_argument('-p', '--file-prefix', dest='prefix', type=str,
        default=prefix, metavar='STR',
        help=HELPTEXT['prefix'] + ' (default: pdb%(default)s)')

    parser.add_argument('-f', '--number-format', dest='numformat', type=str,
        default=DEFAULTS['numformat'], metavar='STR',
        help=HELPTEXT['numformat'] + ' (default: %(default)s)')

    parser.add_argument('-d', '--delimiter', dest='numdelim', type=str,
        default=DEFAULTS['numdelim'], metavar='STR',
        help=HELPTEXT['numdelim'] + ' (default: "%(default)s")')

    parser.add_argument('-x', '--extension', dest='numext', type=str,
        default=DEFAULTS['numext'], metavar='STR',
        help=HELPTEXT['numext'] + ' (default: %(default)s)')

    return parser


def addNMAFigures(parser):

    parser = parser.add_argument_group('figures')

    parser.add_argument('-A', '--all-figures', dest='figall',
        action='store_true',
        default=DEFAULTS['figall'], help=HELPTEXT['figall'])

    parser.add_argument('-R', '--cross-correlations-figure', dest='figcc',
        action='store_true', default=DEFAULTS['figcc'],
        help=HELPTEXT['figcc'])

    parser.add_argument('-Q', '--square-fluctuations-figure', dest='figsf',
        action='store_true', default=DEFAULTS['figsf'],
        help=HELPTEXT['figsf'])

    return parser


def addNMAFigureOptions(parser):

    parser = parser.add_argument_group('figure options')

    parser.add_argument('-F', '--figure-format', dest='figformat', type=str,
        default=DEFAULTS['figformat'], metavar='STR',
        help=DEFAULTS['figformat'] + ' (default: %(default)s)',
        choices=set(FIGFORMATS))

    parser.add_argument('-D', '--dpi', dest='figdpi', type=int,
        default=DEFAULTS['figdpi'], metavar='INT',
        help=HELPTEXT['figdpi'] + ' (default: %(default)s)')

    parser.add_argument('-W', '--width', dest='figwidth', type=float,
        default=DEFAULTS['figwidth'], metavar='FLOAT',
        help=HELPTEXT['figwidth'] + ' (default: %(default)s)')

    parser.add_argument('-H', '--height', dest='figheight', type=float,
        default=DEFAULTS['figheight'], metavar='FLOAT',
        help=HELPTEXT['figheight'] + ' (default: %(default)s)')

    return parser



# -*- coding: utf-8 -*-
import numpy as np

from .flags import PLANTERS
from .atom import Atom
from .fields import ATOMIC_FIELDS, READONLY
from .fields import wrapGetMethod, wrapSetMethod
from .pointer import AtomPointer
from prody import LOGGER

__all__ = ['AtomSubset']


class AtomSubset(AtomPointer):

    """A class for manipulating subset of atoms in an :class:`.AtomGroup`.
    Derived classes are:

      * :class:`.Selection`
      * :class:`.Segment`
      * :class:`.Chain`
      * :class:`.Residue`

    This class stores a reference to an :class:`.AtomGroup` instance, a set of
    atom indices, and active coordinate set index for the atom group."""

    __slots__ = ['_ag', '_indices', '_acsi', '_selstr']

    def __init__(self, ag, indices, acsi, **kwargs):

        AtomPointer.__init__(self, ag, acsi)

        if not isinstance(indices, np.ndarray):
            indices = np.array(indices, int)
        elif not np.issubdtype(indices.dtype, np.signedinteger):
            indices = list(indices[0])
            indices = np.array(indices, int)

        if kwargs.get('unique'):
            self._indices = indices
        else:
            self._indices = np.unique(indices)

        self._selstr = kwargs.get('selstr')

    def __len__(self):

        return len(self._indices)

    def __getitem__(self, index):
        try:
            index = self._indices[index]
            return self._ag[index]
        except:
            raise TypeError('invalid index')

    def getCoords(self):
        """Returns a copy of coordinates from the active coordinate set."""

        if self._ag._coords is not None:
            # Since this is not slicing, a view is not returned
            return self._ag._coords[self.getACSIndex(), self._indices]

    _getCoords = getCoords

    def getAnisous(self):
        """Returns a copy of anisotropic temperature factors from the active coordinate set."""

        if self._ag._anisous is not None:
            # Since this is not slicing, a view is not returned
            return self._ag._anisous[self.getACSIndex(), self._indices]

    _getAnisous = getAnisous

    def setCoords(self, coords):
        """Set coordinates in the active coordinate set."""

        if self._ag._coords is not None:
            self._ag._coords[self.getACSIndex(), self._indices] = coords
            self._ag._setTimeStamp(self.getACSIndex())

    def setAnisous(self, anisous):
        """Set anisotropic temperature factors in the active coordinate set."""

        if self._ag._anisous is not None:
            self._ag._anisous[self.getACSIndex(), self._indices] = anisous
            self._ag._setTimeStamp(self.getACSIndex())

    def getCoordsets(self, indices=None):
        """Returns coordinate set(s) at given *indices*, which may be an integer
        or a list/array of integers."""

        if self._ag._coords is None:
            return None
        if indices is None:
            return self._ag._coords[:, self._indices]
        if isinstance(indices, (int, slice)):
            return self._ag._coords[indices, self._indices]
        if isinstance(indices, (list, np.ndarray)):
            return self._ag._coords[indices, self._indices]
        raise IndexError('indices must be an integer, a list/array of '
                         'integers, a slice, or None')

    _getCoordsets = getCoordsets

    def iterCoordsets(self):
        """Yield copies of coordinate sets."""

        coords = self._ag._getCoordsets()
        if coords is not None:
            indices = self._indices
            for xyz in coords:
                yield xyz[indices]

    _iterCoordsets = iterCoordsets

    def getIndices(self):
        """Returns a copy of the indices of atoms."""

        return self._indices.copy()

    def _getIndices(self):
        """Returns indices of atoms."""

        return self._indices

    def numAtoms(self, flag=None):
        """Returns number of atoms, or number of atoms with given *flag*."""

        return len(self._getSubset(flag)) if flag else len(self._indices)

    def iterAtoms(self):
        """Yield atoms."""

        ag = self._ag
        acsi = self.getACSIndex()
        for index in self._indices:
            yield Atom(ag=ag, index=index, acsi=acsi)

    __iter__ = iterAtoms

    def getData(self, label):
        """Returns a copy of data associated with *label*, if it is present."""

        data = self._ag._getData(label)
        if data is not None:
            return data[self._indices]

    _getData = getData

    def setData(self, label, data):
        """Update *data* associated with *label*.

        :raise AttributeError: when *label* is not in use or read-only"""

        if label in READONLY:
            raise AttributeError('{0} is read-only'.format(repr(label)))
        if label in ATOMIC_FIELDS:
            getattr(self, 'set' + ATOMIC_FIELDS[label].meth_pl)(data)
        else:
            try:
                self._ag._data[label][self._indices] = data
            except KeyError:
                raise AttributeError('data with label {0} must be set for '
                                     'AtomGroup first'.format(repr(label)))

    def getFlags(self, label):
        """Returns a copy of atom flags for given *label*, or **None** when
        flags for *label* is not set."""

        return self._ag._getFlags(label)[self._indices]

    def setFlags(self, label, value):
        """Update flag associated with *label*.

         :raise AttributeError: when *label* is not in use or read-only"""

        if label in PLANTERS:
            raise AttributeError('flag {0} cannot be changed by user'
                                    .format(repr(label)))
        flags = self._ag._getFlags(label)
        if flags is None:
            raise AttributeError('flags with label {0} must be set for '
                                    'AtomGroup first'.format(repr(label)))
        flags[self._indices] = value


for fname, field in ATOMIC_FIELDS.items():

    if field.private:
        continue

    meth = field.meth_pl
    getMeth = 'get' + meth
    setMeth = 'set' + meth
    # Define public method for retrieving a copy of data array
    def getData(self, meth=field.meth_pl, call=field.call):
        data = getattr(self._ag, '_get' + meth)()
        if data is not None:
            return data[self._indices]
    getData = wrapGetMethod(getData)
    getData.__name__ = getMeth
    getData.__doc__ = field.getDocstr('get')
    setattr(AtomSubset, getMeth, getData)
    setattr(AtomSubset, '_' + getMeth, getData)

    if field.readonly:
        continue

    # Define public method for setting values in data array
    def setData(self, value, var=fname, none=field.none):
        array = self._ag._data[var]
        if array is None:
            raise AttributeError(var + ' data is not set')
        array[self._indices] = value
        if none: self._ag._none(none)
    setData = wrapSetMethod(setData)
    setData.__name__ = setMeth
    setData.__doc__ = field.getDocstr('set')
    setattr(AtomSubset, setMeth, setData)

del getData
del setData



"""ProDy is a package for Protein Dynamics, Sequence, and Structure Analysis"""

__version__ = '2.6.1'
__release__ = __version__ # + '-dev' # comment out '-dev' before a release

import sys
import warnings

if sys.version_info[:2] < (2, 7):
    sys.stderr.write('Python 2.6 and older is not supported\n')

if sys.version_info[0] == 3:
    if sys.version_info[1] < 10:
        sys.stderr.write('Python 3.9 and older is not supported\n')
        sys.exit()

try:
    import numpy as np
except ImportError:
    raise ImportError('Numpy is a required package for ProDy')
else:
    if tuple(map(int, np.__version__.split('.')[:2])) < (1, 10):
        raise ImportError('Numpy v1.10 or later is required for ProDy')

try:
    import scipy
except ImportError:
    raise ImportError('Scipy is a required package for ProDy')

DEPRECATION_WARNINGS = False


def deprecate(dep, alt, ver=None, sl=3):
    """Issue a deprecation warning for *dep* and recommend using *alt*."""

    if ver is None:
        ver = list(__version__.split('.')[:2])
        ver[1] = str(int(ver[1]) + 1)
        ver = '.'.join(ver)

    warnings.warn('`{0:s}` is deprecated for removal in v{1:s}, use `{2:s}`.'
                  .format(dep, ver, alt), DeprecationWarning, stacklevel=sl)


def turnonDeprecationWarnings(action='always'):
    """Turn on deprecation warnings for the current session.  By default
     (``action='always'``), deprecation warnings will be printed every time
     a function is called to help identification of multiple occurrence
     of deprecated function and method names.  When ``action='default'``
     is passed, warning will be issued at the first call of a function.
     The latter behavior will automatically kick in when v0.9 is released.
     Until v0.9 is released, restarting the session will turn of warnings.
     This function must be called as ``prody.turnonDeprecationWarnings``."""

    global DEPRECATION_WARNINGS
    DEPRECATION_WARNINGS = True
    warnings.filterwarnings(action, category=DeprecationWarning)


_PY3K = PY3K = sys.version_info[0] > 2
PY2K = not PY3K

__all__ = ['checkUpdates', 'confProDy', 'startLogfile', 'closeLogfile', 'plog']

from . import utilities
from .utilities import *
from .utilities import LOGGER, PackageSettings
from .utilities import getPackagePath, joinRepr, tabulate
__all__.extend(utilities.__all__)
__all__.append('utilities')

SETTINGS = PackageSettings('prody', logger=LOGGER)
SETTINGS.load()

from . import kdtree
from .kdtree import *
__all__.extend(kdtree.__all__)
__all__.append('kdtree')

from . import atomic
from .atomic import *
__all__.extend(atomic.__all__)
__all__.append('atomic')

from .atomic import SELECT

from . import proteins
from .proteins import *
__all__.extend(proteins.__all__)
__all__.append('proteins')

from . import compounds
from .compounds import *
__all__.extend(compounds.__all__)
__all__.append('compounds')

from . import measure
from .measure import *
__all__.extend(measure.__all__)
__all__.append('measure')

from . import database
from .database import *
__all__.extend(database.__all__)
__all__.append('database')

from . import sequence
from .sequence import *
__all__.extend(sequence.__all__)
__all__.append('sequence')

from . import dynamics
from .dynamics import *
__all__.extend(dynamics.__all__)
__all__.append('dynamics')

from . import ensemble
from .ensemble import *
__all__.extend(ensemble.__all__)
__all__.append('ensemble')

from . import trajectory
from .trajectory import *
__all__.extend(trajectory.__all__)
__all__.append('trajectory')

from . import chromatin
from .chromatin import *
__all__.extend(chromatin.__all__)
__all__.append('chromatin')

from . import domain_decomposition
from .domain_decomposition import *
__all__.extend(domain_decomposition.__all__)
__all__.append('domain_decomposition')

#from . import comd
#from .comd import *
#__all__.extend(comd.__all__)
#__all__.append('comd')

import prody
__all__.append('prody')

# default, acceptable values, setter
CONFIGURATION = {
    'backup': (False, None, None),
    'backup_ext': ('.BAK', None, None),
    'auto_show': (False, None, None),
    'ligand_xml_save': (False, None, None),
    'typo_warnings': (True, None, None),
    'check_updates': (0, None, None),
    'auto_secondary': (False, None, None),
    'auto_bonds': (False, None, None),
    'selection_warning': (True, None, None),
    'verbosity': ('debug', list(utilities.LOGGING_LEVELS),
                  LOGGER._setverbosity),
    'pdb_mirror_path': ('', None, proteins.pathPDBMirror),
    'local_pdb_folder': ('', None, proteins.pathPDBFolder),
}


def confProDy(*args, **kwargs):
    """Configure ProDy."""

    if args:
        values = []
        for option in args:
            try:
                values.append(SETTINGS[option])
            except KeyError:
                raise KeyError('{0:s} is not a valid configuration option'
                               .format(repr(option)))
        if len(values) == 1:
            return values[0]
        else:
            return values

    for option, value in kwargs.items():
        try:
            default, acceptable, setter = CONFIGURATION[option]
        except KeyError:
            raise KeyError('{0:s} is not a valid configuration option'
                           .format(repr(option)))
        else:
            try:
                value = type(default)(value)
            except ValueError:
                raise TypeError('{0:s} must be a {1:s}'
                                .format(option, type(default).__name__))
            if acceptable is not None and value not in acceptable:
                raise ValueError('{0:s} must be one of {1:s}'.format(option,
                                 joinRepr(acceptable, sort=True,
                                          last=', or ')))

            SETTINGS[option] = value
            LOGGER.info('ProDy is configured: {0:s}={1:s}'
                        .format(option, repr(value)))
            SETTINGS.save()
            if setter is not None:
                setter(value)

_keys = list(CONFIGURATION)
_keys.sort()
_vals = []
for _key in _keys:
    default, acceptable, setter = CONFIGURATION[_key]
    try:
        if not setter.func_name.startswith('_'):
            seealso = ' See also :func:`.' + setter.func_name + '`.'
    except AttributeError:
        seealso = ''

    if acceptable is None:
        _vals.append(repr(default) + seealso)
    else:
        _vals.append(repr(default) + ' (' +
                     joinRepr(acceptable, sort=True, last=', or ') + ')' +
                     seealso)
    if _key not in SETTINGS:
        SETTINGS[_key] = default

LOGGER._setverbosity(confProDy('verbosity'))

confProDy.__doc__ += '\n\n' + tabulate(['Option'] + _keys,
                                       ['Default (acceptable values)'] + _vals
                                       ) + """

Usage example::

    confProDy('backup')
    confProDy('backup', 'backup_ext')
    confProDy(backup=True, backup_ext='.bak')
    confProDy(backup_ext='.BAK')"""


def plog(*text):
    """Log *text* using ProDy logger at log level info.  Multiple arguments
    are accepted.  Each argument will be converted to string and joined using
    a white space as delimiter."""

    LOGGER.info(' '.join([str(s) for s in text]))


def startLogfile(filename, **kwargs):

    LOGGER.start(filename, **kwargs)

startLogfile.__doc__ = LOGGER.start.__doc__


def closeLogfile(filename):
    """Close logfile with *filename*."""

    LOGGER.close(filename)


def checkUpdates():
    """Check PyPI to see if there is a newer ProDy version available.  Setting
    ProDy configuration parameter *check_updates* to a positive integer will
    make ProDy automatically check updates, e.g.::

      confProDy(check_updates=7) # check at most once a week
      confProDy(check_updates=0) # do not auto check updates
      confProDy(check_updates=-1) # check at the start of every session"""

    pypi_url = 'https://pypi.python.org/pypi'
    if PY3K:
        import xmlrpc.client
        pypi = xmlrpc.client.Server(pypi_url)
    else:
        import xmlrpclib
        pypi = xmlrpclib.Server(pypi_url)
    releases = pypi.package_releases('ProDy')
    if releases[0] == __version__:
        LOGGER.info('You are using the latest ProDy release (v{0:s}).'
                    .format(__version__))
    else:
        LOGGER.info('ProDy v{0:s} is available, you are using {1:s}.'
                    .format(releases[0], __version__))
    if SETTINGS['check_updates']:
        import time
        SETTINGS['last_check'] = time.time()
        SETTINGS.save()


if SETTINGS['check_updates']:

    if SETTINGS.get('last_check') is None:
        SETTINGS['last_check'] = 0
    import time
    if ((time.time() - SETTINGS.get('last_check')) / 3600 / 24 >
            SETTINGS['check_updates']):
        LOGGER.info('Checking PyPI for ProDy updates:')
        checkUpdates()



# -*- coding: utf-8 -*-
"""This module defines functions for calculating physical properties from normal
modes."""

import time

import numpy as np

from prody import LOGGER
from prody.atomic import Atomic
from prody.ensemble import Ensemble, Conformation
from prody.trajectory import TrajBase
from prody.utilities import importLA, checkCoords, div0
from numpy import sqrt, arange, log, polyfit, array

from .nma import NMA
from .modeset import ModeSet
from .mode import VectorBase, Mode, Vector
from .gnm import GNMBase

__all__ = ['calcCollectivity', 'calcCovariance', 'calcCrossCorr',
           'calcFractVariance', 'calcSqFlucts', 'calcRMSFlucts',
           'calcMostMobileNodes', 'calcTempFactors',
           'calcProjection', 'calcCrossProjection',
           'calcSpecDimension', 'calcPairDeformationDist',
           'calcDistFlucts', 'calcHinges', 'calcHitTime',
           'calcAnisousFromModel', 'calcScipionScore', 'calcHemnmaScore']
           #'calcEntropyTransfer', 'calcOverallNetEntropyTransfer']

def calcCollectivity(mode, masses=None, is3d=None):
    """Returns collectivity of the mode.  This function implements collectivity
    as defined in equation 5 of [BR95]_.  If *masses* are provided, they will
    be incorporated in the calculation.  Otherwise, atoms are assumed to have
    uniform masses.

    .. [BR95] Bruschweiler R. Collective protein dynamics and nuclear
       spin relaxation. *J Chem Phys* **1995** 102:3396-3403.

    :arg mode: mode(s) or vector(s)
    :type mode: :class:`.Mode`, :class:`.Vector`, :class:`.ModeSet`,
        :class:`.NMA`, :class:`~numpy.ndarray`

    :arg masses: atomic masses
    :type masses: :class:`numpy.ndarray`
    
    :arg is3d: whether mode is 3d. Default is **None** which means determine 
        the value based on ``mode.is3d()``.
    :type is3d: bool
    """

    if isinstance(mode, np.ndarray):
        V = mode
        ndim = V.ndim
        shape = V.shape

        if is3d is None:
            is3d = False

        if ndim == 0:
            raise ValueError('mode cannot be an empty array')
        elif ndim == 1:
            V = V[:, np.newaxis]

        n = shape[0]
        if is3d:
            n_atoms = n // 3
        else:
            n_atoms = n
    else:
        V, W, is3d_, n_atoms = _getModeProperties(mode)
        if is3d is None:
            is3d = is3d_
    
    colls = []

    def log0(a):
        return log(a + np.finfo(float).eps)

    for v in V.T:
        if is3d:
            u2in = (v ** 2)
            u2in_Nx3 = np.reshape(u2in, (n_atoms, 3))
            u2in = u2in_Nx3.sum(axis=1)
        else:
            u2in = (v ** 2)
        if masses is not None:
            if len(masses) != n_atoms:
                raise ValueError('length of masses must be equal to number of atoms')
            u2in = u2in / masses
        u2in = u2in * (1 / u2in.sum() ** 0.5)
        coll = np.exp(-(u2in * log0(u2in)).sum()) / n_atoms
        colls.append(coll)
    
    if len(colls) == 1:
        return coll
    else:
        return np.array(colls)

def calcSpecDimension(mode):

    """
    :arg mode: mode or vector
    :type mode: :class:`.Mode` or :class:`.Vector`

    """
    # if not isinstance(mode, Mode):
    #     raise TypeError('mode must be a Mode instance')
    
    length = mode.shape[0]
    numbers = arange(2,length+1)
    ds,p=polyfit(log(sqrt(mode[0:int(length*0.25)])),log(numbers[0:int(length*0.25)]),1)
    
    return ds

def calcFracDimension(mode):
    """
    :arg mode: mode or vector
    :type mode: mode or vector """




def calcFractVariance(mode):
    """Returns fraction of variance explained by the *mode*.  Fraction of
    variance is the ratio of the variance along a mode to the trace of the
    covariance matrix of the model."""

    if isinstance(mode, Mode):
        var = mode.getVariance()
        trace = mode.getModel()._getTrace()
    elif isinstance(mode, (ModeSet, NMA)):
        var = mode.getVariances()
        if isinstance(mode, ModeSet):
            trace = mode.getModel()._getTrace()
        else:
            trace = mode._getTrace()
    else:
        raise TypeError('mode must be a Mode instance')
    if trace is None:
        raise ValueError('modes are not calculated')

    return var / trace


def calcProjection(ensemble, modes, rmsd=True, norm=False):
    """Returns projection of conformational deviations onto given modes.
    *ensemble* coordinates are used to calculate the deviations that are
    projected onto *modes*.  For K conformations and M modes, a (K,M)
    matrix is returned.

    :arg ensemble: an ensemble, trajectory or a conformation for which
        deviation(s) will be projected, or a deformation vector
    :type ensemble: :class:`.Ensemble`, :class:`.Conformation`,
        :class:`.Vector`, :class:`.Trajectory`
        
    :arg modes: up to three normal modes
    :type modes: :class:`.Mode`, :class:`.ModeSet`, :class:`.NMA`

    By default, root-mean-square deviation (RMSD) along the normal mode is
    calculated. To calculate the raw projection pass ``rmsd=False``.

    By default, the projection is not normalized. If you would like it to be,
    pass ``norm=True``.

    :class:`.Vector` instances are accepted as *ensemble* argument to allow
    for projecting a deformation vector onto normal modes."""

    if not isinstance(ensemble, (Ensemble, Conformation, Vector, TrajBase)):
        raise TypeError('ensemble must be Ensemble, Conformation, Vector, '
                        'or a TrajBase, not {0}'.format(type(ensemble)))
    if not isinstance(modes, (NMA, ModeSet, VectorBase)):
        raise TypeError('rows must be NMA, ModeSet, or Mode, not {0}'
                        .format(type(modes)))
    if not modes.is3d():
        raise ValueError('modes must be 3-dimensional')
    if isinstance(ensemble, Vector):
        n_atoms = ensemble.numAtoms()
    else:
        n_atoms = ensemble.numSelected()
    if n_atoms != modes.numAtoms():
        raise ValueError('number of atoms are not the same')
    if isinstance(ensemble, Vector):
        if not ensemble.is3d():
            raise ValueError('ensemble must be a 3d vector instance')
        deviations = ensemble._getArray()
    elif isinstance(ensemble, (Ensemble, Conformation)):
        deviations = ensemble.getDeviations()
    else:
        nfi = ensemble.nextIndex()
        ensemble.goto(0)
        deviations = np.array([frame.getDeviations() for frame in ensemble])
        ensemble.goto(nfi)
    if deviations.ndim == 3:
        deviations = deviations.reshape((deviations.shape[0],
                                         deviations.shape[1] * 3))
    elif deviations.ndim == 2:
        deviations = deviations.reshape((1, deviations.shape[0] * 3))
    else:
        deviations = deviations.reshape((1, deviations.shape[0]))
    la = importLA()
    if norm:
        N = la.norm(deviations)
        if N != 0:
            deviations = deviations / N
    projection = np.dot(deviations, modes._getArray())
    if rmsd:
        projection = (1 / (n_atoms ** 0.5)) * projection
    return projection


def calcCrossProjection(ensemble, mode1, mode2, scale=None, **kwargs):
    """Returns projection of conformational deviations onto modes from
    different models.

    :arg ensemble: ensemble for which deviations will be projected
    :type ensemble: :class:`.Ensemble`

    :arg mode1: normal mode to project conformations onto
    :type mode1: :class:`.Mode`, :class:`.Vector`

    :arg mode2: normal mode to project conformations onto
    :type mode2: :class:`.Mode`, :class:`.Vector`

    :arg scale: scale width of the projection onto mode1 (``x``) or mode2(``y``),
        an optimized scaling factor (scalar) will be calculated by default 
        or a value of scalar can be passed.
        
    This function uses calcProjection and its arguments can be 
    passed to it as keyword arguments.
    By default, this function applies RMSD scaling and normalisation. 
    These can be turned off with ``rmsd=False`` and ``norm=False``."""

    if not isinstance(ensemble, (Ensemble, Conformation, Vector, TrajBase)):
        raise TypeError('ensemble must be Ensemble, Conformation, Vector, '
                        'or a Trajectory, not {0}'.format(type(ensemble)))
    if not isinstance(mode1, VectorBase):
        raise TypeError('mode1 must be a Mode instance, not {0}'
                        .format(type(mode1)))
    if not mode1.is3d():
        raise ValueError('mode1 must be 3-dimensional')
    if not isinstance(mode2, VectorBase):
        raise TypeError('mode2 must be a Mode instance, not {0}'
                        .format(type(mode2)))
    if not mode2.is3d():
        raise ValueError('mode2 must be 3-dimensional')

    if scale is not None:
        assert isinstance(scale, str), 'scale must be a string'
        scale = scale.lower()
        assert scale in ('x', 'y'), 'scale must be x or y'

    xcoords = calcProjection(ensemble, mode1, kwargs.get('rmsd', True), kwargs.get('norm', True))
    ycoords = calcProjection(ensemble, mode2, kwargs.pop('rmsd', True), kwargs.pop('norm', True))
    if scale:
        scalar = kwargs.get('scalar', None)
        if scalar:
            assert isinstance(scalar, (float, int)), 'scalar must be a number'
        else:
            scalar = ((ycoords.max() - ycoords.min()) /
                      (xcoords.max() - xcoords.min())
                      ) * np.sign(np.dot(xcoords, ycoords))
            if scale == 'x':
                LOGGER.info('Projection onto {0} is scaled by {1:.2f}'
                            .format(mode1, scalar))
            else:
                scalar = 1 / scalar
                LOGGER.info('Projection onto {0} is scaled by {1:.2f}'
                            .format(mode2, scalar))

        if scale == 'x':
            xcoords = xcoords * scalar
        else:
            ycoords = ycoords * scalar

    return xcoords, ycoords

def _getModeProperties(modes):
    V = []; W = []; is3d = None; n_atoms = 0
    if isinstance(modes, VectorBase):
        V = modes._getArray()
        if isinstance(modes, Mode):
            W = modes.getVariance()
        else:
            W = 1.
        V = np.asarray([V]).T
        W = np.asarray([[W]])
        is3d = modes.is3d()
        n_atoms = modes.numAtoms()
    elif isinstance(modes, (NMA, ModeSet)):
        V = modes._getArray()
        W = np.diag(modes.getVariances())
        is3d = modes.is3d()
        n_atoms = modes.numAtoms()
    elif isinstance(modes, list):
        for mode in modes:
            if not isinstance(mode, VectorBase):
                raise TypeError('modes can be a list of VectorBase instances, '
                                'not {0}'.format(type(mode)))
            V.append(mode._getArray())
            if isinstance(mode, Mode):
                W.append(modes.getVariance())
            else:
                W.append(1.)
            if is3d is None:
                is3d = mode.is3d()
                n_atoms = mode.numAtoms()
            else:
                if is3d != mode.is3d():
                    raise ValueError('modes must be either all from ANM or GNM')
                if n_atoms != mode.numAtoms():
                    raise ValueError('each mode in the list must have the same number of atoms')
        V = np.array(V).T
        W = np.diag(W)
    else:
        raise TypeError('modes must be a Mode, NMA, ModeSet instance, '
                        'or a list of Mode instances, not {0}'.format(type(modes)))
    return V, W, is3d, n_atoms

def calcSqFlucts(modes):
    """Returns sum of square-fluctuations for given set of normal *modes*.
    Square fluctuations for a single mode is obtained by multiplying the
    square of the mode array with the variance (:meth:`.Mode.getVariance`)
    along the mode.  For :class:`.PCA` and :class:`.EDA` models built using
    coordinate data in Å, unit of square-fluctuations is |A2|, for
    :class:`.ANM` and :class:`.GNM`, on the other hand, it is arbitrary or
    relative units."""

    V = []; W = []; is3d = None; n_atoms = 0
    V, W, is3d, n_atoms = _getModeProperties(modes)

    sq_flucts = np.dot(V * V, W).sum(axis=1)

    if is3d:
        sq_flucts_Nx3 = np.reshape(sq_flucts, (n_atoms, 3))
        sq_flucts = sq_flucts_Nx3.sum(axis=1)
    return sq_flucts

def calcRMSFlucts(modes):
    """Returns root mean square fluctuation(s) (RMSF) for given set of normal *modes*.
    This is calculated just by doing the square root of the square fluctuations """
    sq_flucts = calcSqFlucts(modes)
    if len(np.where(sq_flucts<0)[0]) != 0:
        raise ValueError("Square Fluctuation should not contain negative values, please check input modes")

    return sq_flucts ** 0.5

def calcMostMobileNodes(modes, **kwargs):
    """Returns indices for nodes with highest root mean square fluctuations (RMSFs) for given set of normal *modes*
    above a particular *percentile* and/or *cutoff*.

    :arg percentile: percentile for internal cutoff (between 0 and 100).
        Default 0 takes all values
    :type percentile: int

    :arg cutoff: user-defined cutoff, default is to take all values
    :type cutoff: float
    """
    rmsf = calcRMSFlucts(modes)

    cutoff = kwargs.get('cutoff', rmsf.min())
    inds = np.nonzero(rmsf > cutoff)[0]

    percentile = kwargs.get('percentile', 0)
    cutoff = np.percentile(rmsf, percentile)
    inds = inds[np.nonzero(rmsf[inds] > cutoff)[0]]

    return inds

def calcCrossCorr(modes, n_cpu=1, norm=True):
    """Returns cross-correlations matrix.  For a 3-d model, cross-correlations
    matrix is an NxN matrix, where N is the number of atoms.  Each element of
    this matrix is the trace of the submatrix corresponding to a pair of atoms.
    Cross-correlations matrix may be calculated using all modes or a subset of modes
    of an NMA instance.  For large systems, calculation of cross-correlations
    matrix may be time consuming.  Optionally, multiple processors may be
    employed to perform calculations by passing ``n_cpu=2`` or more."""

    if not isinstance(n_cpu, int):
        raise TypeError('n_cpu must be an integer')
    elif n_cpu < 1:
        raise ValueError('n_cpu must be equal to or greater than 1')

    if not isinstance(modes, (Mode, Vector, NMA, ModeSet)):
        if isinstance(modes, list):
            try:
                is3d = modes[0].is3d()
            except:
                raise TypeError('modes must be a list of Mode or Vector instances, '
                            'not {0}'.format(type(modes)))
        else:
            raise TypeError('modes must be a Mode, Vector, NMA, or ModeSet instance, '
                            'not {0}'.format(type(modes)))
    else:
        is3d = modes.is3d()

    if is3d:
        model = modes
        if isinstance(modes, (Mode, ModeSet)):
            model = modes._model
            if isinstance(modes, (Mode)):
                indices = [modes.getIndex()]
                n_modes = 1
            else:
                indices = modes.getIndices()
                n_modes = len(modes)
        elif isinstance(modes, Vector):
                indices = [0]
                n_modes = 1
        else:
            n_modes = len(modes)
            indices = np.arange(n_modes)
            
        array = model._getArray()
        n_atoms = model._n_atoms

        if not isinstance(modes, Vector):
            variances = model._vars
        else:
            array = array.reshape(-1, 1)
            variances = np.ones(1)

        if n_cpu == 1:
            s = (n_modes, n_atoms, 3)
            arvar = (array[:, indices]*variances[indices]).T.reshape(s)
            array = array[:, indices].T.reshape(s)
            covariance = np.tensordot(array.transpose(2, 0, 1),
                                      arvar.transpose(0, 2, 1),
                                      axes=([0, 1], [1, 0]))
        else:
            import multiprocessing
            n_cpu = min(multiprocessing.cpu_count(), n_cpu)
            queue = multiprocessing.Queue()
            size = n_modes // n_cpu
            for i in range(n_cpu):
                if n_cpu - i == 1:
                    indices = modes.getIndices()[i*size:]
                else:
                    indices = modes.getIndices()[i*size:(i+1)*size]
                process = multiprocessing.Process(
                    target=_crossCorrelations,
                    args=(queue, n_atoms, array, variances, indices))
                process.start()
            while queue.qsize() < n_cpu:
                time.sleep(0.05)
            covariance = queue.get()
            while queue.qsize() > 0:
                covariance += queue.get()
    else:
        covariance = calcCovariance(modes)
    if norm:
        diag = np.power(covariance.diagonal(), 0.5)
        D = np.outer(diag, diag)
        covariance = div0(covariance, D)
    return covariance


def _crossCorrelations(queue, n_atoms, array, variances, indices):
    """Calculate covariance-matrix for a subset of modes."""

    n_modes = len(indices)
    arvar = (array[:, indices] * variances[indices]).T.reshape((n_modes,
                                                                n_atoms, 3))
    array = array[:, indices].T.reshape((n_modes, n_atoms, 3))
    covariance = np.tensordot(array.transpose(2, 0, 1),
                              arvar.transpose(0, 2, 1),
                              axes=([0, 1], [1, 0]))
    queue.put(covariance)

def calcDistFlucts(modes, n_cpu=1, norm=True):
    """Returns the matrix of distance fluctuations (i.e. an NxN matrix
    where N is the number of residues, of MSFs in the inter-residue distances)
    computed from the cross-correlation matrix (see Eq. 12.E.1 in [IB18]_). 
    The arguments are the same as in :meth:`.calcCrossCorr`.

    .. [IB18] Dill K, Jernigan RL, Bahar I. Protein Actions: Principles and
       Modeling. *Garland Science* **2017**. """

    cc = calcCrossCorr(modes, n_cpu=n_cpu, norm=norm)
    cc_diag = np.diag(cc).reshape(-1,1)
    distFluct = cc_diag.T + cc_diag -2.*cc
    return distFluct

def calcTempFactors(modes, atoms):
    """Returns temperature (β) factors calculated using *modes* from a
    :class:`.ANM` or :class:`.GNM` instance scaled according to the 
    experimental B-factors from *atoms*."""

    model = modes.getModel()
    if not isinstance(model, GNMBase):
        raise TypeError('modes must come from GNM or ANM')
    if model.numAtoms() != atoms.numAtoms():
        raise ValueError('modes and atoms must have same number of nodes')
    sqf = calcSqFlucts(modes)
    expBetas = atoms.getBetas()
    # add warning message if experimental B-factors are zeros or meaningless (e.g., having same values)?
    if expBetas.max() < 0.5 or expBetas.std() < 0.5:
        LOGGER.warning('Experimental B-factors are quite small or meaningless. The calculated B-factors may be incorrect.')
    return sqf * (expBetas.sum() / sqf.sum())


def calcCovariance(modes):
    """Returns covariance matrix calculated for given *modes*.
    This is 3Nx3N for 3-d models and NxN (equivalent to cross-correlations) 
    for 1-d models such as GNM."""

    if isinstance(modes, NMA):
        return modes.getCovariance()
    else:
        V, W, _, _ = _getModeProperties(modes)
        return np.dot(V, np.dot(W, V.T))


def calcPairDeformationDist(model, coords, ind1, ind2, kbt=1.):                                       
    """Returns distribution of the deformations in the distance contributed by each mode 
    for selected pair of residues *ind1* *ind2* using *model* from a :class:`.ANM`.
    Method described in [EB08]_ equation (10) and figure (2).     
    
    .. [EB08] Eyal E., Bahar I. Toward a Molecular Understanding of 
        the Anisotropic Response of Proteins to External Forces:
        Insights from Elastic Network Models. *Biophys J* **2008** 94:3424-34355. 
    
    :arg model: this is an 3-dimensional :class:`NMA` instance from a :class:`.ANM`
        calculations.
    :type model: :class:`.ANM`  

    :arg coords: a coordinate set or an object with :meth:`getCoords` method.
        Recommended: ``coords = parsePDB('pdbfile').select('protein and name CA')``.
    :type coords: :class:`~numpy.ndarray`.

    :arg ind1: first residue number.
    :type ind1: int 
    
    :arg ind2: second residue number.
    :type ind2: int 
    """

    try:
        resnum_list = coords.getResnums()
        resnam_list = coords.getResnames()
        coords = (coords._getCoords() if hasattr(coords, '_getCoords') else
                coords.getCoords())
    except AttributeError:
        try:
            checkCoords(coords)
        except TypeError:
            raise TypeError('coords must be a Numpy array or an object '
                            'with `getCoords` method')
    
    if not isinstance(model, NMA):
        raise TypeError('model must be a NMA instance')
    elif not model.is3d():
        raise TypeError('model must be a 3-dimensional NMA instance')
    elif len(model) == 0:
        raise ValueError('model must have normal modes calculated')
    
    linalg = importLA()
    n_atoms = model.numAtoms()
    n_modes = model.numModes()
    LOGGER.timeit('_pairdef')

    r_ij = np.zeros((n_atoms,n_atoms,3))
    r_ij_norm = np.zeros((n_atoms,n_atoms,3))

    for i in range(n_atoms):
        for j in range(i+1,n_atoms):
            r_ij[i][j] = coords[j,:] - coords[i,:]
            r_ij[j][i] = r_ij[i][j]
            r_ij_norm[i][j] = r_ij[i][j]/linalg.norm(r_ij[i][j])
            r_ij_norm[j][i] = r_ij_norm[i][j]

    eigvecs = model.getEigvecs()
    eigvals = model.getEigvals()
    
    D_pair_k = []
    mode_nr = []
    ind1 = ind1 - resnum_list[0]
    ind2 = ind2 - resnum_list[0]

    for m in range(6,n_modes):
        U_ij_k = [(eigvecs[ind1*3][m] - eigvecs[ind2*3][m]), (eigvecs[ind1*3+1][m] \
            - eigvecs[ind2*3+1][m]), (eigvecs[ind1*3+2][m] - eigvecs[ind2*3+2][m])] 
        D_ij_k = abs(sqrt(kbt/eigvals[m])*(np.vdot(r_ij_norm[ind1][ind2], U_ij_k)))  
        D_pair_k.append(D_ij_k)
        mode_nr.append(m)

    LOGGER.report('Deformation was calculated in %.2lfs.', label='_pairdef')
    
    return mode_nr, D_pair_k

def calcHinges(modes, atoms=None, flag=False):
    """Returns the hinge sites identified using normal modes.     

    :arg modes: normal modes of which will be used to identify hinge sites
    :type modes: :class:`.GNM`  
    
    :arg atoms: an Atomic object on which to map hinges. The output will then be a selection. 
    :type atoms: :class:`.Atomic`

    :arg flag: whether return flag or index array. Default is **False**
    :type flag: bool

    """

    def identify(v):
        # obtain the signs of eigenvector
        s = np.sign(v)
        # obtain the relative magnitude of eigenvector
        mag = np.sign(np.diff(np.abs(v)))
        # obtain the cross-overs
        torf = np.diff(s)!=0
        torf = np.append(torf, [False], axis=0)
        # find which side is more close to zero
        for j, m in enumerate(mag):
            if torf[j] and m < 0:
                torf[j+1] = True
                torf[j] = False

        return torf

    if modes.is3d():
        raise ValueError('3D models are not supported.')

    # obtain the eigenvectors
    V = modes.getArray()
    if V.ndim == 1:
        hinges = identify(V)
    elif V.ndim == 2:
        _, n = V.shape
        hinges = []
        for i in range(n):
            v = V[:, i]
            torf = identify(v)
            hinges.append(torf)

        hinges = np.stack(hinges).T
    else:
        raise TypeError('wrong dimension of the array: %d'%V.ndim)
    
    if not flag:
        hinge_list = np.where(hinges)[0]
        if atoms is not None:
            if isinstance(atoms, Atomic):
                return atoms[hinge_list]
            else:
                raise TypeError('atoms should be an Atomic object')
        return sorted(set(hinge_list))
    return hinges

def calcHitTime(model, method='standard'):
    """Returns the hit and commute times between pairs of nodes calculated 
    based on a :class:`.NMA` object. 

    .. [CB95] Chennubhotla C., Bahar I. Signal Propagation in Proteins and Relation
    to Equilibrium Fluctuations. *PLoS Comput Biol* **2007** 3(9).

    :arg model: model to be used to calculate hit times
    :type model: :class:`.NMA`  

    :arg method: method to be used to calculate hit times. Available options are 
        ``"standard"`` or ``"kirchhoff"``. Default is ``"standard"``
    :type method: str

    :returns: (:class:`~numpy.ndarray`, :class:`~numpy.ndarray`)
    """

    try:
        K = model.getKirchhoff()
    except AttributeError:
        raise TypeError('model must be an NMA instance')

    if K is None:
        raise ValueError('model not built')
    
    method = method.lower()

    D = np.diag(K)
    A = np.diag(D) - K

    start = time.time()
    linalg = importLA()
    if method == 'standard':
        st = D / sum(D)

        P = np.dot(np.diag(D**(-1)), A)
        W = np.ones((len(st), 1)) * st.T
        Z = linalg.pinv(np.eye(P.shape[0], P.shape[1]) - P + W)

        H = np.ones((len(st), 1)) * np.diag(Z).T - Z
        H = H / W
        H = H.T

    elif method == 'kirchhoff':
        K_inv = linalg.pinv(K)
        sum_D = sum(D)

        T1 = (sum_D * np.ones((len(D),1)) * np.diag(K_inv)).T

        T2 = sum_D * K_inv
        T3_i = np.dot((np.ones((len(D),1)) * D), K_inv)

        H = T1 - T2 + T3_i - T3_i.T

    C = H + H.T

    LOGGER.debug('Hit and commute times are calculated in  {0:.2f}s.'
                 .format(time.time()-start)) 
    return H, C


def calcAnisousFromModel(model, ):
    """Returns a Nx6 matrix containing anisotropic B factors (ANISOU lines)
    from a covariance matrix calculated from **model**.

    :arg model: 3D model from which to calculate covariance matrix
    :type model: :class:`.ANM`, :class:`.PCA`

    .. ipython:: python

       from prody import *
       protein = parsePDB('1ejg')
       anm, calphas = calcANM(protein)
       adp_matrix = calcAnisousFromModel(anm)"""

    if not isinstance(model, (NMA, Mode)) or not model.is3d():
        raise TypeError('model must be of type ANM, PCA or Mode, not {0}'
                        .format(type(model)))

    cov = calcCovariance(model)
    n_atoms = model.numAtoms()
    
    submatrices = [cov[i*3:(i+1)*3, i*3:(i+1)*3] for i in range(n_atoms)]

    anisou = np.zeros((n_atoms, 6))
    for index, submatrix in enumerate(submatrices):
        anisou[index, 0] = submatrix[0, 0]
        anisou[index, 1] = submatrix[1, 1]
        anisou[index, 2] = submatrix[2, 2]
        anisou[index, 3] = submatrix[0, 1]
        anisou[index, 4] = submatrix[0, 2]
        anisou[index, 5] = submatrix[1, 2]
    return anisou


def calcScipionScore(modes):
    """Calculate the score from hybrid electron microscopy normal mode analysis (HEMNMA) 
    [CS14]_ as implemented in the Scipion continuousflex plugin [MH20]_. This score 
    prioritises modes as a function of mode number and collectivity order.

    .. [CS14] Sorzano COS, de la Rosa-Trevín JM, Tama F, Jonić S.
       Hybrid Electron Microscopy Normal Mode Analysis graphical interface and protocol.
       *J Struct Biol* **2014** 188:134-41.

    .. [MH20] Harastani M, Sorzano COS, Jonić S. 
       Hybrid Electron Microscopy Normal Mode Analysis with Scipion.
       *Protein Sci* **2020** 29:223-236.

    :arg modes: mode(s) or vector(s)
    :type modes: :class:`.Mode`, :class:`.Vector`, :class:`.ModeSet`, :class:`.NMA`
    """
    n_modes = modes.numModes()
    
    if n_modes > 1:
        collectivityList = list(calcCollectivity(modes))
    else:
        collectivityList = [calcCollectivity(modes)]

    idxSorted = [i[0] for i in sorted(enumerate(collectivityList),
                                      key=lambda x: x[1],
                                      reverse=True)]

    score = np.zeros(n_modes)
    modeNum = list(range(n_modes))

    for i in range(n_modes):
        score[idxSorted[i]] = idxSorted[i] + modeNum[i] + 2  

    score = score / (2.0 * n_modes) 

    return score

calcHemnmaScore = calcScipionScore



"""This module contains unit tests for :mod:`~prody.dynamics`."""

import numpy as np
from numpy import arange
from numpy.testing import *

from prody.utilities import importDec
dec = importDec()

from prody import *
from prody import LOGGER
from prody.tests import unittest
from prody.tests.datafiles import *

LOGGER.verbosity = 'none'

ATOL = 1e-5
RTOL = 0


ATOMS = parseDatafile('1ubi_ca')
COORDS = ATOMS.getCoords()

ANM_HESSIAN = parseDatafile('anm1ubi_hessian', symmetric=True)
ANM_EVALUES = parseDatafile('anm1ubi_evalues')[:,1].flatten()
ANM_EVECTORS = parseDatafile('anm1ubi_vectors')[:,1:]

GNM_KIRCHHOFF = parseDatafile('gnm1ubi_kirchhoff', symmetric=True, skiprows=1)
GNM_EVALUES = parseDatafile('gnm1ubi_evalues')[:,1].flatten()
GNM_EVECTORS = parseDatafile('gnm1ubi_vectors', usecols=arange(3,23))


anm = ANM()
anm.buildHessian(ATOMS)
anm.calcModes(n_modes=None, zeros=True)


gnm = GNM()
gnm.buildKirchhoff(ATOMS)
gnm.calcModes(n_modes=None, zeros=True)

ATOMS2 = parseDatafile('2gb1_truncated')
rtb = RTB()
rtb.buildHessian(ATOMS2, ATOMS2.getBetas().astype(int))
RTB_HESSIAN = parseDatafile('rtb2gb1_hessian', symmetric=True)
RTB_PROJECT = parseDatafile('rtb2gb1_project')

HITTIME = parseDatafile('hit1ubi')
COMMUTETIME = parseDatafile('commute1ubi')


ATOMS3 = parseDatafile('2nwl')
COORDS3 = ATOMS3.ca.getCoords()

EXANM_HESSIAN = parseDatafile('exanm2nwl_hessian.coo', symmetric=True)
EXANM_EVALUES = parseDatafile('exanm2nwl_evalues.dat')[:,1].flatten()
EXANM_EVECTORS = parseDatafile('exanm2nwl_vectors.dat')[:,1:]

exanm = ANM()
exanm.buildHessian(COORDS3)
exanm.calcModes(n_modes=None, zeros=True)

class testGNMBase(unittest.TestCase):

    def setUp(self):

        self.model = gnm


    def testGetCutoff(self):
        """Testing return type of :meth:`~.GNMBase.getCutoff`."""

        self.assertIsInstance(self.model.getCutoff(), float,
                              'getCutoff failed to return a float')

    def testGetGamma(self):
        """Testing return type of :meth:`~.GNMBase.getGamma`."""

        self.assertIsInstance(self.model.getGamma(), float,
                              'getCutoff failed to return a float')


class TestANMResults(testGNMBase):

    def testEigenvalues(self):
        """Test eigenvalues."""

        assert_allclose(anm[:len(ANM_EVALUES)].getEigvals(), ANM_EVALUES,
                        rtol=RTOL, atol=ATOL*10,
                        err_msg='failed to get correct eigenvalues')


    def testEigenvectors(self):
        """Test eigenvectors."""

        _temp = np.abs(np.dot(anm[6:6+ANM_EVECTORS.shape[1]].getEigvecs().T,
                              ANM_EVECTORS))
        assert_allclose(_temp, np.eye(20), rtol=RTOL, atol=ATOL,
                        err_msg='failed to get correct eigenvectors')

    def testHessian(self):
        """Test Hessian matrix."""

        assert_allclose(anm.getHessian(), ANM_HESSIAN, rtol=0, atol=ATOL,
                        err_msg='failed to get correct Hessian matrix')


    def testVariances(self):
        """Test variances."""

        assert_allclose(anm[6:len(ANM_EVALUES)].getVariances(),
                        1/ANM_EVALUES[6:],
                        rtol=0, atol=ATOL*100,
                        err_msg='failed to get correct variances')

    def testGetHessian(self):
        assert_equal(anm.getHessian(), anm._getHessian(),
                     err_msg='failed to _get correct Hessian matrix')

    def testHessianSymmetry(self):
        hessian = anm._getHessian()
        assert_equal(hessian, hessian.T, 'hessian is not symmetric')

    def testHessianSums(self):
        hessian = anm._getHessian()
        zeros = np.zeros(hessian.shape[0])
        assert_allclose(hessian.sum(0), zeros,
                        rtol=0, atol=ATOL,
                        err_msg='hessian columns do not add up to zero')
        assert_allclose(hessian.sum(1), zeros,
                        rtol=0, atol=ATOL,
                        err_msg='hessian rows do not add up to zero')
        
    def testModeSlicing(self):
        indices1 = slice(227, 230)
        assert_equal(anm[indices1].numModes(), 1,
                         'slicing ANM for last mode does not work')
        indices = slice(228, 230)
        self.assertRaises(IndexError, anm.__getitem__, indices)

class TestEXANMResults(testGNMBase):

    def testEigenvalues(self):
        """Test eigenvalues."""

        assert_allclose(exanm[:len(EXANM_EVALUES)].getEigvals(), EXANM_EVALUES,
                        rtol=RTOL, atol=ATOL*10,
                        err_msg='failed to get correct eigenvalues')


    def testEigenvectors(self):
        """Test eigenvectors."""

        _temp = np.abs(np.dot(exanm[6:6+EXANM_EVECTORS.shape[1]].getEigvecs().T,
                              EXANM_EVECTORS))
        assert_allclose(_temp, np.eye(20), rtol=RTOL, atol=ATOL,
                        err_msg='failed to get correct eigenvectors')

    def testHessian(self):
        """Test Hessian matrix."""

        assert_allclose(exanm.getHessian(), EXANM_HESSIAN, rtol=0, atol=ATOL,
                        err_msg='failed to get correct Hessian matrix')


    def testVariances(self):
        """Test variances."""

        assert_allclose(exanm[6:len(EXANM_EVALUES)].getVariances(),
                        1/EXANM_EVALUES[6:],
                        rtol=0, atol=ATOL*100,
                        err_msg='failed to get correct variances')

    def testGetHessian(self):
        assert_equal(exanm.getHessian(), exanm._getHessian(),
                     err_msg='failed to _get correct Hessian matrix')

    def testHessianSymmetry(self):
        hessian = exanm._getHessian()
        assert_equal(hessian, hessian.T, 'hessian is not symmetric')

    def testHessianSums(self):
        hessian = exanm._getHessian()
        zeros = np.zeros(hessian.shape[0])
        assert_allclose(hessian.sum(0), zeros,
                        rtol=0, atol=ATOL,
                        err_msg='hessian columns do not add up to zero')
        assert_allclose(hessian.sum(1), zeros,
                        rtol=0, atol=ATOL,
                        err_msg='hessian rows do not add up to zero')

'''
class TestANMSparse(unittest.TestCase):

    """Test result from using sparse matrices."""

    @dec.slow
    @unittest.skipIf(True, 'not completed')
    def testSparse(self):

        anm = ANM()
        anm.buildHessian(COORDS, sparse=True)
        assert_allclose(anm.getHessian().toarray(), ANM_HESSIAN,
                        rtol=0, atol=ATOL,
                        err_msg='failed to get correct sparse Hessian matrix')
        anm.calcModes(None)
        assert_allclose(anm[:len(ANM_EVALUES)].getEigvals(), ANM_EVALUES,
                        rtol=RTOL, atol=ATOL*10,
                        err_msg='failed to get correct eigenvalues')
        _temp = np.abs(np.dot(anm[6:6+ANM_EVECTORS.shape[1]].getEigvecs().T,
                              ANM_EVECTORS))
        assert_allclose(_temp, np.eye(20), rtol=RTOL, atol=ATOL,
                        err_msg='failed to get correct eigenvectors')
'''

class TestGNMResults(testGNMBase):

    def testEigenvalues(self):
        assert_allclose(gnm[:21].getEigvals(), GNM_EVALUES[:21],
                        rtol=RTOL, atol=ATOL*100,
                        err_msg='failed to get correct slow eigenvalues')

        assert_allclose(gnm[-21:].getEigvals(), GNM_EVALUES[21:],
                        rtol=RTOL, atol=ATOL*100,
                        err_msg='failed to get correct fast eigenvalues')

    def testEigenvectors(self):
        _temp = np.abs(np.dot(gnm[1:21].getEigvecs().T, GNM_EVECTORS))
        assert_allclose(_temp, np.eye(20), rtol=RTOL, atol=ATOL*10,
                       err_msg='failed to get correct eigenvectors')

    def testKirchhoff(self):
        assert_allclose(gnm._getKirchhoff(), GNM_KIRCHHOFF,
                        rtol=0, atol=ATOL,
                        err_msg='failed to get correct Kirchhoff matrix')

    def testGetKirchoff(self):
        assert_equal(gnm.getKirchhoff(), gnm._getKirchhoff(),
                     err_msg='failed to _get correct Kirchhoff matrix')

    def testKirchhoffSymmetry(self):
        kirchhoff = gnm._getKirchhoff()
        assert_equal(kirchhoff, kirchhoff.T, 'kirchhoff is not symmetric')

    def testKirchhoffSums(self):
        kirchhoff = gnm._getKirchhoff()
        zeros = np.zeros(kirchhoff.shape[0])
        assert_equal(kirchhoff.sum(0), zeros,
                     'kirchhoff columns do not add up to zero')
        assert_equal(kirchhoff.sum(1), zeros,
                     'kirchhoff rows do not add up to zero')

    def testBuildKirchoffSlow(self):
        slow = GNM()
        slow.buildKirchhoff(ATOMS)
        assert_equal(slow._getKirchhoff(), gnm._getKirchhoff(),
                     'slow method does not reproduce same Kirchhoff')

    def testCommuteTime(self):
        hitTime, commuteTime = calcHitTime(gnm)

        assert_allclose(hitTime, HITTIME,
                        rtol=0, atol=ATOL,
                        err_msg='failed to get correct hit times')
        assert_allclose(commuteTime, COMMUTETIME,
                        rtol=0, atol=ATOL,
                        err_msg='failed to get correct commute times')             

class TestGNM(unittest.TestCase):

    def setUp(self):

        self.model = GNM()
        self.buildMatrix = self.model.buildKirchhoff
        self.setMatrix = self.model.setKirchhoff
        self.getMatrix = self.model.getKirchhoff
        self.getExpected = gnm.getKirchhoff

    def testBuildMatrixCoordsWrongType(self):
        """Test response to wrong type *coords* argument."""

        self.assertRaises(TypeError, self.buildMatrix, 'nogood')

    def testBuildMatrixWrongCoords(self):
        """Test response to wrong coords.dtype."""

        array = np.array([['a','a','a'] for i in range(10)])
        self.assertRaises(ValueError, self.buildMatrix, array)

    def testBuildMatrixCoordsArray(self):
        """Test output when  *coords* is an array."""

        self.buildMatrix(COORDS)
        assert_equal(self.getMatrix(), self.getExpected(),
                     'failed to get correct matrix')

    def testBuildMatrixCutoffWrongType(self):
        """Test response to cutoff of wrong type."""

        self.assertRaises(TypeError, self.buildMatrix, COORDS, 'none')

    def testBuildMatrixCutoffInvalidValue(self):
        """Test response to wrong *cutoff* argument."""

        self.assertRaises(ValueError, self.buildMatrix, COORDS, -1)

    def testBuildMatrixInvalidGamma(self):
        """Test response to invalid *gamma* argument."""

        self.assertRaises(TypeError, self.buildMatrix, COORDS, gamma='none')

    def testBuildMatrixWrongGamma(self):
        """Test response to wrong *gamma* argument."""

        self.assertRaises(ValueError, self.buildMatrix, COORDS, gamma=0)

    def testSetMatrixWrongType(self):
        """Test response to wrong matrix type argument."""

        self.assertRaises(TypeError, self.setMatrix, list(np.ones((3,3))))

    def testSetMatrixWrongDim(self):
        """Test response to wrong dim *kirchhoff* argument."""

        self.assertRaises(ValueError, self.setMatrix, np.ones((3,4,3)))

    def testSetMatrixNonSquare(self):
        """Test response to non-square matrix."""

        self.assertRaises(ValueError, self.setMatrix, np.ones((3,4)))

    def testSetMatrixWrongDtype(self):
        """Test response to wrong matrix.dtype."""

        array = np.array([['a','a','a'] for i in range(3)])
        self.assertRaises(ValueError, self.setMatrix, array)

    def testSetMatrixAcceptableDtype(self):
        """Test response to acceptable matrix.dtype."""

        self.assertIsNone(self.setMatrix(np.ones((30,30), int)),
                          'failed to set an acceptable array')

class TestANM(TestGNM):

    def setUp(self):

        self.model = ANM()
        self.anm = self.model
        self.buildMatrix = self.model.buildHessian
        self.getMatrix = self.model.getHessian
        self.setMatrix = self.model.setHessian
        self.getExpected = anm.getHessian

    def testSetHessianWrongShape(self):
        """Test response to wrong shape *hessian* argument."""

        self.assertRaises(ValueError, self.model.setHessian, np.ones((5,5)))

    def testBuildHessianSlow(self):
        slow = ANM()
        slow.buildHessian(ATOMS, slow=True)
        assert_allclose(slow._getHessian(), anm._getHessian(),
                        rtol=0, atol=ATOL,
                        err_msg='slow method does not reproduce same Hessian')
        assert_equal(slow._getKirchhoff(), anm._getKirchhoff(),
                     'slow method does not reproduce same Kirchhoff')
    

class TestGNMCalcModes(unittest.TestCase):

    def setUp(self):
        pass

class TestRTB(unittest.TestCase):

    def testHessian(self):

        assert_allclose(RTB_HESSIAN, rtb._getHessian(),
                        rtol=0, atol=ATOL,
                        err_msg='expected Hessian is not produced')

    def testProjection(self):

        assert_allclose(RTB_PROJECT, rtb._getProjection(),
                        rtol=0, atol=ATOL,
                        err_msg='expected projection matrix is not produced')


    def testCalcModes(self):

        rtb.calcModes()

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""This module defines classes and functions to parse and write Hi-C data files, 
visualize, and perform elastic network analysis on Hi-C data.

Parse/write Hi-C files
=====================

Following ProDy functions are for parsing and writing Hi-C files:

  * :func:`.parseHiC` - parse Hi-C data file
  * :func:`.parseHiCStream` - parse Hi-C data stream
  * :func:`.writeMap` - write Hi-C data to text file

Visualize Hi-C data
=====================

Following ProDy functions are for visualizing Hi-C data:

  * :func:`.showMap` - show Hi-C contact map
  * :func:`.showDomains` - show Hi-C structural domains

Save/load HiC class
-------------------

    * :func:`.saveHiC`
    * :func:`.loadHiC`

"""

import prody
LOGGER = prody.LOGGER
SETTINGS = prody.SETTINGS

__all__ = []

from . import hic
from .hic import *
__all__.extend(hic.__all__)

from . import norm
from .norm import *
__all__.extend(norm.__all__)

from . import cluster
from .cluster import *
__all__.extend(cluster.__all__)

from . import functions
from .functions import *
__all__.extend(functions.__all__)



"""This module contains unit tests for :mod:`~prody.proteins`."""

import os

from numpy.testing import *

from prody.utilities import importDec
dec = importDec()

from prody import *
from prody import LOGGER
from prody.tests import TEMPDIR, unittest
from prody.tests.datafiles import *

LOGGER.verbosity = 'none'


class TestHTTP(unittest.TestCase):

    def setUp(self):

        self.pdb = ['1ubi', '1aar', 'arg', 1234]
        self.fns = []
        self.len = [683, 1218, None, None]
        self.fetch = fetchPDBviaHTTP
        self.protocol = 'HTTP'


    @dec.slow
    def testCompressed(self):

        self.fns = self.fetch(*self.pdb, folder=TEMPDIR)

        for fn, pdb, n_atoms in zip(self.fns, self.pdb, self.len):
            if fn is None:
                continue
            self.assertTrue(os.path.isfile(fn),
                'failed to fetch PDB file via ' + self.protocol)
            atoms = parsePDB(fn)
            self.assertIsNotNone(atoms, 'PDB file ({}) empty, download via {}'
                                        .format(fn, self.protocol))
            self.assertEqual(n_atoms, len(atoms))


    @dec.slow
    def testDecompressed(self):

        self.fns = self.fetch(*self.pdb, folder=TEMPDIR, compressed=False)

        for fn, pdb, n_atoms in zip(self.fns, self.pdb, self.len):
            if fn is None:
                continue
            self.assertTrue(os.path.isfile(fn),
                'failed to fetch PDB file via ' + self.protocol)
            atoms = parsePDB(fn)
            self.assertIsNotNone(atoms, 'PDB file ({}) empty, download via {}'
                                        .format(fn, self.protocol))
            self.assertEqual(n_atoms, len(atoms))


    def tearDown(self):


        for fn in self.fns:
            if fn is None:
                continue
            try:
                #os.remove(fn)
                pass
            except:
                pass



# -*- coding: utf-8 -*-
"""This module defines :class:`Crossterm` for dealing with crossterm information provided
by using :meth:`.AtomGroup.setCrossterms` method."""

from numbers import Integral
import numpy as np

RAD2DEG = 180 / np.pi

__all__ = ['Crossterm']


class Crossterm(object):

    """A pointer class for crosstermd atoms.  Following built-in functions are
    customized for this class:

    * :func:`len` returns crossterm length, i.e. :meth:`getSize`
    * :func:`iter` yields :class:`~.Atom` instances"""

    __slots__ = ['_ag', '_acsi', '_indices']

    def __init__(self, ag, indices, acsi=None):

        self._ag = ag
        self._indices = np.array(indices)
        if acsi is None:
            self._acsi = ag.getACSIndex()
        else:
            self._acsi = acsi

    def __repr__(self):

        one, two, three, four = self._indices
        names = self._ag._getNames()
        return '<Crossterm: {0}({1})--{2}({3})--{4}({5}--{6}({7})) from {8}>'.format(
            names[one], one, names[two], two,
            names[three], three, names[four], four,
            str(self._ag))

    def __str__(self):

        one, two, three, four = self._indices
        names = self._ag._getNames()
        return '{0}({1})--{2}({3})--{4}({5})--{6}({7})'.format(
            names[one], one, names[two], two,
            names[three], three, names[four], four)

    def __eq__(self, other):

        return (isinstance(other, Crossterm) and other.getAtomGroup() is self._ag
                and (np.all(other.getIndices() == self._indices) or
                     np.all(other.getIndices() == list(reversed(self._indices)))))

    def __ne__(self, other):

        return not self.__eq__(other)

    def __size__(self):

        return self.getSize()

    def __iter__(self):

        for index in self._indices:
            yield self._ag[index]

    def getAtomGroup(self):
        """Returns atom group."""

        return self._ag

    def getAtoms(self):
        """Returns crosstermd atoms."""

        return (self._ag[self._indices[0]], self._ag[self._indices[1]], self._ag[self._indices[2]])

    def getIndices(self):
        """Returns indices of crosstermd atoms."""

        return self._indices.copy()

    def getSize(self, radian=False):
        """Returns crossterm size."""

        a1, a2, a3 = self.getVectors()

        v1 = np.cross(a1, a2)
        v1 = v1 / (v1 * v1).sum(-1)**0.5
        v2 = np.cross(a2, a3)
        v2 = v2 / (v2 * v2).sum(-1)**0.5
        porm = np.sign((v1 * a3).sum(-1))
        rad = np.arccos((v1*v2).sum(-1) / ((v1**2).sum(-1) * (v2**2).sum(-1))**0.5)
        if not porm == 0:
            rad = rad * porm
        if radian:
            return rad
        else:
            return rad * RAD2DEG

    def getVectors(self):
        """Returns bond vectors that originate from the central atom."""

        one, two, three, four = self._indices
        acsi = self.getACSIndex()
        vector1 = self._ag._coords[acsi, two] - self._ag._coords[acsi, one]
        vector2 = self._ag._coords[acsi, three] - self._ag._coords[acsi, two]
        vector3 = self._ag._coords[acsi, four] - self._ag._coords[acsi, three]
        return vector1, vector2, vector3

    def getACSIndex(self):
        """Returns index of the coordinate set."""

        acsi = self._acsi
        if acsi >= self._ag._n_csets:
            raise ValueError('{0} has fewer coordsets than assumed by {1}'
                             .format(str(self._ag), str(self)))
        return acsi

    def setACSIndex(self, index):
        """Set the coordinate set at *index* active."""

        if self._ag._coords is None:
            raise AttributeError('coordinates are not set')

        if not isinstance(index, Integral):
            raise TypeError('index must be an integer')

        n_csets = self._ag._n_csets
        if n_csets <= index or n_csets < abs(index):
            raise IndexError('coordinate set index is out of range')

        if index < 0:
            index += n_csets

        self._acsi = index


def evalCrossterms(crossterms, n_atoms):
    """Returns an array mapping atoms to their crosstermd neighbors and an array
    that stores number of crossterms made by each atom."""

    numcrossterms = np.bincount(crossterms.reshape((crossterms.shape[0] * 4)))
    dmap = np.zeros((n_atoms, numcrossterms.max(), 3), int)
    dmap.fill(-1)
    index = np.zeros(n_atoms, int)
    for crossterm in crossterms:
        a, b, c, d = crossterm
        dmap[a, index[a]] = [b, c, d]
        dmap[b, index[b]] = [a, c, d]
        dmap[c, index[c]] = [a, b, d]
        dmap[d, index[d]] = [a, b, c]
        index[crossterm] += 1
    return dmap, numcrossterms


def trimCrossterms(crossterms, indices):
    """Returns crossterms between atoms at given indices."""

    iset = set(indices)
    crossterms = [crossterm for crossterm in crossterms if crossterm[0]
                 in iset and crossterm[1] in iset and crossterm[2] in iset and crossterm[3] in iset]
    if crossterms:
        newindices = np.zeros(indices.max()+1, int)
        newindices[indices] = np.arange(len(indices))
        return newindices[np.array(crossterms)]



# -*- coding: utf-8 -*-

"""This module provides the the WatFinder toolkit that detects, predicts and analyzes water bridges.
"""

__author__ = 'Karolina Mikulska-Ruminska'
__credits__ = ['Frane Doljanin', 'Karolina Mikulska-Ruminska']
__email__ = ['karolamik@fizyka.umk.pl', 'fdoljanin@pmfst.hr']

import multiprocessing as mp

from numbers import Number
import numpy as np
import os

from itertools import combinations
from collections import deque

from prody import PY3K
if PY3K:
    from enum import Enum, auto
else:
    from enum import Enum
    import enum
    from itertools import count
    def auto(it=count()):
        return next(it)
    enum.auto=auto

from copy import copy

from prody import LOGGER, SETTINGS
from prody.atomic import Atom, Atomic, AtomGroup
from prody.atomic.flags import DEFAULTS
from prody.ensemble import Ensemble
from prody.measure import calcAngle, calcDistance
from prody.measure.contacts import findNeighbors
from prody.proteins import writePDB, parsePDB

from prody.utilities import showFigure, showMatrix


__all__ = ['calcWaterBridges', 'calcWaterBridgesTrajectory', 'getWaterBridgesInfoOutput',
           'calcWaterBridgesStatistics', 'getWaterBridgeStatInfo',
           'calcWaterBridgeMatrix', 'showWaterBridgeMatrix',
           'calcBridgingResiduesHistogram', 'calcWaterBridgesDistribution',
           'savePDBWaterBridges', 'savePDBWaterBridgesTrajectory',
           'saveWaterBridges', 'parseWaterBridges', 'findClusterCenters',
           'filterStructuresWithoutWater', 'selectSurroundingsBox',
           'findCommonSelectionTraj']


class ResType(Enum):
    WATER = auto()
    PROTEIN = auto()
    ION = auto()


class AtomNode:
    def __init__(self, atom, type):
        self.atom = atom
        self.hydrogens = []
        self.type = type
        self.bonds = []
        self.bondedAtoms = []
        self.isVisited = False


class RelationList:
    def __init__(self, numNodes):
        self.nodes = [None for _ in range(numNodes)]

    def resetVisits(self, atomsToReset=[]):
        if atomsToReset:
            for atom in atomsToReset:
                self[atom].isVisited = False

            return

        for node in self.nodes:
            if node:
                node.isVisited = False

    def addNode(self, atom, type):
        self[atom.getIndex()] = AtomNode(atom, type)

        return self[atom]

    def addHBond(self, bond):
        donor, acceptor = bond.donor, bond.acceptor
        self[donor].bonds.append(bond)
        self[acceptor].bonds.append(bond)

        if acceptor.getIndex() not in map(lambda a: a.getIndex(), self[donor].bondedAtoms):
            self[donor].bondedAtoms.append(acceptor)
            self[acceptor].bondedAtoms.append(donor)

    def removeUnbonded(self):
        for i, node in enumerate(self.nodes):
            if node and not node.bonds:
                self[i] = None

    def __getitem__(self, key):
        if isinstance(key, Atom):
            key = key.getIndex()

        return self.nodes[key]

    def __setitem__(self, key, value):
        self.nodes[key] = value


class HBondConstraints:
    def __init__(self, donors, acceptors, angleWW=None, anglePAWD=None, anglePDWA=None):
        self.donors = donors
        self.acceptors = acceptors
        self.angleWW = angleWW
        self.anglePAWD = anglePAWD
        self.anglePDWA = anglePDWA

    def addAngles(self, angleWW, anglePAWD, anglePDWA):
        self.angleWW = angleWW
        self.anglePAWD = anglePAWD
        self.anglePDWA = anglePDWA


class HydrogenBond:
    def __init__(self, donor, acceptor):
        self.donor = donor
        self.acceptor = acceptor

    @staticmethod
    def checkIsHBond(donor, acceptor, constraints):
        if donor.type == ResType.ION or acceptor.type == ResType.ION:
            return False
        if donor.type != ResType.WATER and donor.atom.getName()[0] not in constraints.donors:
            return False
        if acceptor.type != ResType.WATER and acceptor.atom.getName()[0] not in constraints.acceptors:
            return False

        anglesForComb = {
            (ResType.WATER, ResType.WATER): constraints.angleWW,
            (ResType.WATER, ResType.PROTEIN): constraints.anglePAWD,
            (ResType.PROTEIN, ResType.WATER): constraints.anglePDWA,
        }

        angleRange = anglesForComb[(donor.type, acceptor.type)]
        if not angleRange:
            return True

        return HydrogenBond.checkIsHBondAngle(donor, acceptor, angleRange)

    @staticmethod
    def checkIsHBondAngle(donor, acceptor, angleRange):
        for hydrogen in donor.hydrogens:
            bondAngle = calcAngle(donor.atom, hydrogen, acceptor.atom)
            if angleRange[0] < bondAngle < angleRange[1]:
                return True

        return False

class CoordinationBond:
    def __init__(self, donor, acceptor):
        self.donor = donor
        self.acceptor = acceptor

    @staticmethod
    def checkIsCoorBond(donor, acceptor, constraints):
        if donor.type != ResType.WATER and donor.atom.getName() not in constraints.donors:
            return False
        if acceptor.type != ResType.WATER and acceptor.atom.getName() not in constraints.acceptors:
            return False
        if donor.type != ResType.ION and acceptor.type != ResType.ION:
            return False

        return True


def calcBridges(relations, hydrophilicList, method, maxDepth, maxNumResidues):
    waterBridges = []

    for atom in hydrophilicList:
        observedNode = relations[atom]
        if not observedNode:
            continue

        newBridges = []
        if method == 'chain':
            newBridges = getBridgeChain_BFS(
                observedNode, relations, maxDepth+1)
            relations.resetVisits()
        else:
            newBridges = getBridgeForResidue_BFS(
                observedNode, relations, maxDepth, maxNumResidues)

        waterBridges += newBridges

    return waterBridges


def getChainBridgeTuple(bridge):
    i_1, i_2 = bridge[0].getIndex(), bridge[-1].getIndex()
    return (min(i_1, i_2), max(i_1, i_2))


def getUniqueElements(list, key):
    unique, uniqueKeys = [], []
    for element in list:
        elementKey = key(element)
        if elementKey in uniqueKeys:
            continue

        unique.append(element)
        uniqueKeys.append(elementKey)

    return unique


def getBridgeChain_BFS(observedNode, relationsList, maxDepth):
    queue = deque([(observedNode.atom, [])])
    waterBridges = []

    while queue:
        currentAtom, currentPath = queue.popleft()
        currentNode = relationsList[currentAtom]

        if currentNode.isVisited:
            continue

        currentNode.isVisited = True

        if currentNode.type == ResType.PROTEIN and len(currentPath):
            waterBridges.append(currentPath + [currentAtom])
            continue

        if len(currentPath) == maxDepth:
            continue

        for atom in currentNode.bondedAtoms:
            queue.append(
                (atom, currentPath + [currentAtom]))

    return waterBridges


def getBridgeForResidue_BFS(observedNode, relationsList, maxDepth, maxNumResidues):
    waterBridges = []

    for waterAtom in observedNode.bondedAtoms:
        bridgeAtoms = []
        newWaters = []
        queue = deque([(waterAtom, 0)])
        while queue:
            currentAtom, depth = queue.popleft()
            currentNode = relationsList[currentAtom]

            if currentNode.isVisited or (maxDepth and depth > maxDepth):
                continue

            bridgeAtoms.append(currentAtom)

            if currentNode.type == ResType.PROTEIN:
                continue
            if currentNode.type == ResType.WATER:
                currentNode.isVisited = True
                newWaters.append(currentNode.atom)

            for atom in currentNode.bondedAtoms:
                queue.append((atom, depth+1))

        bridgeAtomIndices = list(
            set(map(lambda a: a.getIndex(), bridgeAtoms)))

        numProteinResidues = sum(
            relationsList[atomIndex].type == ResType.PROTEIN for atomIndex in bridgeAtomIndices)
        numResidues = len(bridgeAtomIndices)

        if (not maxNumResidues or numResidues <= maxNumResidues) and numProteinResidues >= 2:
            waterBridges.append(bridgeAtomIndices)
        else:
            relationsList.resetVisits(newWaters)

    return waterBridges


class AtomicOutput:
    def __init__(self, proteins, waters):
        self.proteins = proteins
        self.waters = waters


def getInfoOutput(waterBridgesAtomic):
    output = []
    for bridge in waterBridgesAtomic:
        bridgeOutput = []

        for atom in bridge.proteins:
            residueInfo = "{0}{1}".format(atom.getResname(), 
                                          atom.getResnum())
            atomInfo = "{0}_{1}".format(atom.getName(),
                                        atom.getIndex())
            chainInfo = atom.getChid()
            bridgeOutput += [residueInfo, atomInfo, chainInfo]

        for atom_1, atom_2 in combinations(bridge.proteins, r=2):
            bridgeOutput += [calcDistance(atom_1, atom_2)]

        bridgeOutput += [len(bridge.waters)]
        bridgeOutput += [
            list(map(lambda w: "{0}_{1}".format(w.getChid(),
                                                w.getIndex()), 
                     bridge.waters))]

        output.append(bridgeOutput)

    return output


def getWaterBridgesInfoOutput(waterBridgesAtomic):
    """Converts single frame/trajectory atomic output from calcWaterBridges/Trajectory to info output.

    :arg waterBridgesAtomic: water bridges from calcWaterBridges/Trajectory
    :type waterBridgesAtomic: list
    """
    isSingleFrame = isinstance(waterBridgesAtomic[0], AtomicOutput)
    if isSingleFrame:
        return getInfoOutput(waterBridgesAtomic)

    output = []
    for frame in waterBridgesAtomic:
        currentFrameInfo = getInfoOutput(frame)
        output.append(currentFrameInfo)

    return output


def getAtomicOutput(waterBridges, relations):
    output = []
    for bridge in waterBridges:
        proteinAtoms, waterAtoms = [], []
        for atomIndex in bridge:
            if relations[atomIndex].type in [ResType.PROTEIN, ResType.ION]:
                proteinAtoms.append(relations[atomIndex].atom)
            else:
                waterAtoms.append(relations[atomIndex].atom)

        output.append(AtomicOutput(proteinAtoms, waterAtoms))

    return output


def getElementsRegex(elements):
    return '[{0}].*'.format("|".join(elements))


def calcWaterBridges(atoms, **kwargs):
    """Compute water bridges for a protein that has water molecules.

    :arg atoms: Atomic object from which atoms are considered
    :type atoms: :class:`.Atomic`

    :arg method: cluster or chain, where chain find shortest water bridging path between two protein atoms
        default is 'chain'
    :type method: string 'cluster' | 'chain'

    :arg distDA: maximal distance between water/protein donor and acceptor
        default is 3.5
    :type distDA: int, float

    :arg distWR: maximal distance between considered water and any residue
        default is 4
    :type distWR: int, float

    :arg anglePDWA: angle range where protein is donor and water is acceptor
        default is (100, 200)
    :type anglePDWA: (int, int)

    :arg anglePAWD: angle range where protein is acceptor and water is donor
        default is (100, 140)
    :type anglePDWA: (int, int)

    :arg angleWW: angle between water donor/acceptor
        default is (140, 180)
    :type angleWW: (int, int)

    :arg maxDepth: maximum number of waters in chain/depth of residues in cluster
        default is 2
    :type maxDepth: int, None

    :arg maxNumRes: maximum number of water+protein residues in cluster
        default is None
    :type maxNumRes: int, None

    :arg donors: which atoms to count as donors 
        default is ['N', 'O', 'S', 'F']
    :type donors: list

    :arg acceptors: which atoms to count as acceptors 
        default is ['N', 'O', 'S', 'F']
    :type acceptors: list

    :arg output: return information arrays, (protein atoms, water atoms), or just atom indices per bridge
        default is 'atomic'
    :type output: 'info' | 'atomic' | 'indices'

    :arg isInfoLog: should log information
        default is True
    :type output: bool

    :arg selstr: selection string for focusing analysis
        default of **None** focuses on everything
    :type selstr: str

    :arg expand_selection: whether to expand the selection with 
        :func:`.selectSurroundingsBox`, selecting a box surrounding it.
        Default is **False**
    :type expand_selection: bool

    :arg considered_atoms_sel: selection string for which atoms to consider
        Default is **"protein"**
    :type considered_atoms_sel: str
    """

    method = kwargs.pop('method', 'chain')
    distDA = kwargs.pop('distDA', 3.5)
    distWR = kwargs.pop('distWR', 4)
    anglePDWA = kwargs.pop('anglePDWA', (100, 200))
    anglePAWD = kwargs.pop('anglePAWD', (100, 140))
    angleWW = kwargs.pop('angleWW', (140, 180))
    maxDepth = kwargs.pop('maxDepth', 2)
    maxNumResidues = kwargs.pop('maxNumRes', None)
    donors = kwargs.pop('donors', ['N', 'O', 'S', 'F'])
    acceptors = kwargs.pop('acceptors', ['N', 'O', 'S', 'F'])
    outputType = kwargs.pop('output', 'atomic')
    isInfoLog = kwargs.pop('isInfoLog', True)
    DIST_COVALENT_H = 1.4
    prefix = kwargs.pop('prefix', '')
    considered_atoms_sel = kwargs.pop('considered_atoms_sel', "protein")

    if method not in ['chain', 'cluster']:
        raise TypeError('Method should be chain or cluster.')
    if outputType not in ['info', 'atomic', 'indices']:
        raise TypeError('Output can be info, atomic or indices.')

    selstr = kwargs.pop('selstr', None)
    if selstr is not None:
        selection = atoms.select(selstr).copy()

        expand_selection = kwargs.pop('expand_selection', False)
        if expand_selection:
            atoms = selectSurroundingsBox(atoms, selection, **kwargs).copy()
        else:
            atoms = selection.copy()

    water = atoms.select('water')
    if water is None:
        raise ValueError('atoms has no water so cannot be analysed with WatFinder')

    relations = RelationList(len(atoms))
    tooFarAtoms = atoms.select(
        'water and not within {0} of protein'.format(distWR))
    if tooFarAtoms is None:
        consideredAtoms = atoms
    else:
        consideredAtoms = ~tooFarAtoms

    waterHydrogens = consideredAtoms.select('water and hydrogen') or []
    waterOxygens = consideredAtoms.select('water and oxygen')
    waterHydroOxyPairs = findNeighbors(
        waterOxygens, DIST_COVALENT_H, waterHydrogens) if waterHydrogens else []
    for oxygen in waterOxygens:
        relations.addNode(oxygen, ResType.WATER)
    for pair in waterHydroOxyPairs:
        oxygen, hydrogen, _ = pair
        relations[oxygen].hydrogens.append(hydrogen)

    proteinHydrophilic = consideredAtoms.select(
        '({0}) and name "{1}" and within {2} of water'.format(considered_atoms_sel,
            getElementsRegex(set(donors+acceptors)), distWR))

    proteinHydrogens = consideredAtoms.select('({0}) and hydrogen'.format(considered_atoms_sel)) or []
    proteinHydroPairs = findNeighbors(
        proteinHydrophilic, DIST_COVALENT_H, proteinHydrogens) if proteinHydrogens else []
    for hydrophilic in proteinHydrophilic:
        if hydrophilic is not None:
            if hydrophilic.getFlag('ion'):
                relations.addNode(hydrophilic, ResType.ION)
            else:
                relations.addNode(hydrophilic, ResType.PROTEIN)
    for pair in proteinHydroPairs:
        hydrophilic, hydrogen, _ = pair
        relations[hydrophilic].hydrogens.append(hydrogen)

    contactingWaters = findNeighbors(waterOxygens, distDA)
    contactingWaterProtein = findNeighbors(
        waterOxygens, distDA, proteinHydrophilic)

    contactingWaterNodes = list(
        map(lambda ww: (relations[ww[0]], relations[ww[1]]), contactingWaters))
    contactingWaterProteinNodes = list(
        map(lambda wp: (relations[wp[0]], relations[wp[1]]), contactingWaterProtein))

    constraints = HBondConstraints(acceptors, donors)
    if len(waterHydrogens) + len(proteinHydrogens):
        constraints.addAngles(angleWW, anglePAWD, anglePDWA)
    else:
        LOGGER.info('No hydrogens detected, angle criteria will not be used.')

    constraints2 = HBondConstraints(acceptors, donors) # not taking angles

    for pair in contactingWaterNodes + contactingWaterProteinNodes:
        for a, b in [(0, 1), (1, 0)]:
            if HydrogenBond.checkIsHBond(pair[a], pair[b], constraints):
                newHBond = HydrogenBond(pair[a].atom, pair[b].atom)
                relations.addHBond(newHBond)

            elif CoordinationBond.checkIsCoorBond(pair[a], pair[b], constraints2):
                newBond = CoordinationBond(pair[a].atom, pair[b].atom)
                relations.addHBond(newBond)

    relations.removeUnbonded()

    waterBridgesWithIndices = calcBridges(
        relations, proteinHydrophilic, method, maxDepth, maxNumResidues)
    if method == 'chain':
        waterBridgesWithIndices = getUniqueElements(
            waterBridgesWithIndices, getChainBridgeTuple)

    log_string = '{0} water bridges detected using method {1}'.format(
        len(waterBridgesWithIndices), method)
    if prefix != '':
        log_string += ' for ' + prefix
    LOGGER.info(log_string)

    if method == 'atomic':
        LOGGER.info('Call getInfoOutput to convert atomic to info output.')

    atomicOutput = getAtomicOutput(waterBridgesWithIndices, relations)
    infoOutput = getInfoOutput(atomicOutput)

    if isInfoLog:
        for bridge in infoOutput:
            LOGGER.info(' '.join(map(str, bridge)))

    if outputType == 'info':
        return infoOutput
    if outputType == 'atomic':
        return atomicOutput

    return waterBridgesWithIndices


# took from interactions.py
def calcWaterBridgesTrajectory(atoms, trajectory, **kwargs):
    """Computes water bridges for a given trajectory. Kwargs for options are the same as in calcWaterBridges.

    :arg atoms: Atomic object from which atoms are considered
    :type atoms: :class:`.Atomic`

    :arg trajectory: Trajectory data coming from a DCD, ensemble or multi-model PDB file.
    :type trajectory: :class:`.Trajectory', :class:`.Ensemble`, :class:`.Atomic`

    :arg start_frame: frame to start from
    :type start_frame: int

    :arg stop_frame: frame to stop
    :type stop_frame: int

    :arg max_proc: maximum number of processes to use
        default is half of the number of CPUs
    :type max_proc: int

    :arg selstr: selection string for focusing analysis
        default of **None** focuses on everything
    :type selstr: str

    :arg expand_selection: whether to expand the selection with 
        :func:`.selectSurroundingsBox`, selecting a box surrounding it.
        Default is **False**
    :type expand_selection: bool

    If selstr is provided, a common selection will be found across all frames
    combining selections satifying the criteria in each.

    :arg return_selection: whether to return the combined common selection
        Default is **False** to keep expected behaviour.
        However, this output is required when using selstr.
    :type return_selection: bool

    :arg considered_atoms_sel: selection string for which atoms to consider
        Default is **"protein"**
    :type considered_atoms_sel: str  
    """
    start_frame = kwargs.pop('start_frame', 0)
    stop_frame = kwargs.pop('stop_frame', -1)
    max_proc = kwargs.pop('max_proc', mp.cpu_count()//2)
    selstr = kwargs.pop('selstr', None)
    expand_selection = kwargs.pop('expand_selection', False)
    return_selection = kwargs.pop('return_selection', False)
    padding = kwargs.pop('padding', 0)

    if trajectory is not None:
        if isinstance(trajectory, Atomic):
            trajectory = Ensemble(trajectory)

        # nfi = trajectory._nfi
        # trajectory.reset()
        # numFrames = trajectory._n_csets

        if stop_frame == -1:
            traj = trajectory[start_frame:]
        else:
            traj = trajectory[start_frame:stop_frame+1]

        atoms_copy = atoms.copy()
        if selstr is not None:
            selection, indices = findCommonSelectionTraj(atoms, traj, selstr,
                                                         expand_selection=expand_selection,
                                                         return_selection=False,
                                                         padding=padding)
            LOGGER.info('Common selection found with {0} atoms and {1} protein chains'.format(
                selection.numAtoms(), len(list(selection.protein.getHierView()))))

        def analyseFrame(j0, start_frame, frame0, interactions_all):
            LOGGER.info('Frame: {0}'.format(j0))
            atoms_copy = atoms.copy()
            atoms_copy.setCoords(frame0.getCoords())

            if indices is not None:
                atoms_copy = atoms_copy[indices]
                kwargs['selstr'] = atoms_copy.getSelstr()

            interactions = calcWaterBridges(
                atoms_copy, isInfoLog=False, 
                prefix='frame {0}'.format(j0),
                **kwargs)
            interactions_all[j0-start_frame] = interactions

        if max_proc == 1:
            interactions_all = []
            for j0, frame0 in enumerate(traj, start=start_frame):
                interactions_all.append([])
                analyseFrame(j0, start_frame, frame0, interactions_all)
        else:
            with mp.Manager() as manager:
                interactions_all = manager.list()
                for j0, frame0 in enumerate(traj, start=start_frame):
                    interactions_all.append([])

                j0 = start_frame
                while j0 < traj.numConfs()+start_frame:

                    processes = []
                    for _ in range(max_proc):
                        frame0 = traj[j0-start_frame]
                        
                        p = mp.Process(target=analyseFrame, args=(j0, start_frame,
                                                                 frame0,
                                                                 interactions_all))
                        p.start()
                        processes.append(p)

                        j0 += 1
                        if j0 >= traj.numConfs()+start_frame:
                            break

                    for p in processes:
                        p.join()

                interactions_all = interactions_all[:]

        # trajectory._nfi = nfi

    else:
        if atoms.numCoordsets() > 1:
            def analyseFrame(i, interactions_all):
                frameNum = i+start_frame
                LOGGER.info('Model: {0}'.format(frameNum))
                atoms.setACSIndex(i+start_frame)
                interactions = calcWaterBridges(
                    atoms, isInfoLog=False, prefix='frame {0}'.format(frameNum),
                    **kwargs)
                interactions_all[i] = interactions

            if max_proc == 1:
                interactions_all = []
                for i in range(len(atoms.getCoordsets()[start_frame:stop_frame])):
                    interactions_all.append([])
                    analyseFrame(i, interactions_all)
            else:
                with mp.Manager() as manager:
                    interactions_all = manager.list()
                    for i in range(len(atoms.getCoordsets()[start_frame:stop_frame])):
                        interactions_all.append([])

                    i = start_frame
                    while i < len(atoms.getCoordsets()[start_frame:stop_frame]):
                        processes = []
                        for _ in range(max_proc):
                            p = mp.Process(target=analyseFrame, args=(i, interactions_all))
                            p.start()
                            processes.append(p)

                            i += 1
                            if i >= len(atoms.getCoordsets()[start_frame:stop_frame]):
                                break

                        for p in processes:
                            p.join()

                    interactions_all = interactions_all[:]
        else:
            LOGGER.info('Include trajectory or use multi-model PDB file.')

    if return_selection:
        if indices is not None:
            atoms_copy = atoms_copy[indices]
            kwargs['selstr'] = atoms_copy.getSelstr()

        return interactions_all, atoms_copy
    
    return interactions_all


def getResidueName(atom, use_segname=False):
    result = f'{atom.getResname()}{atom.getResnum()}{atom.getChid()}'
    if use_segname:
        result += f'{atom.getSegname()}'

    return result


class DictionaryList:
    def __init__(self, initialValue):
        self.values = {}
        self.initialValue = initialValue

    def __getitem__(self, key):
        if key in self.values:
            return self.values[key]

        return copy(self.initialValue)

    def __setitem__(self, key, value):
        self.values[key] = value

    def keys(self):
        return self.values.keys()

    def removeDuplicateKeys(self, criterion):
        keysCopy = list(self.values.keys())
        for key in keysCopy:
            if criterion(self.values.keys(), key):
                del self.values[key]


def getResInfo(atoms, **kwargs):
    considered_atoms_sel = kwargs.pop('considered_atoms_sel', "protein")
    dict = {}
    nums = atoms.select(considered_atoms_sel).getResnums()
    names = atoms.select(considered_atoms_sel).getResnames()
    chids = atoms.select(considered_atoms_sel).getChids()

    for i, num in enumerate(nums):
        dict[num] = "{0}{1}{2}".format(names[i], num, chids[i])

    return dict


def getWaterBridgeStatInfo(stats, atoms, **kwargs):
    """Converts calcWaterBridgesStatistic indices output to info output from stat.

    :arg stats: statistics returned by calcWaterBridgesStatistics, output='indices'
    :type stats: dictionary

    :arg atoms: Atomic object from which atoms are considered
    :type atoms: :class:`.Atomic`
    """
    residueInfo = getResInfo(atoms, **kwargs)
    infoOutput = {}
    for key, value in stats.items():
        x_id, y_id = key
        x_info, y_info = residueInfo[x_id], residueInfo[y_id]
        newKey = (x_info, y_info)

        infoOutput[newKey] = value

    return infoOutput


def calcWaterBridgesStatistics(frames, trajectory, **kwargs):
    """Returns statistics. Value is percentage of bridge appearance of frames for each residue.

    :arg frames: list of water bridges from calcWaterBridgesTrajectory(), output='atomic'
    :type frames: list

    :arg output: return dictorinary whose keys are tuples of resnames or resids
        default is 'indices'
    :type output: 'info' | 'indices'

    :arg filename: name of file to save statistic information if wanted
        default is None
    :type filename: string

    :arg considered_atoms_sel: selection string for which atoms to consider
        Default is **"protein"**
    :type considered_atoms_sel: str
    """
    output = kwargs.pop('output', 'indices')
    filename = kwargs.pop('filename', None)
    if output not in ['info', 'indices']:
        raise TypeError('Output should be info or indices!')

    considered_atoms_sel = kwargs.pop('considered_atoms_sel', None)
    if considered_atoms_sel is not None:
        trajectory.select(considered_atoms_sel)

    allCoordinates = trajectory.getCoordsets()
    interactionCount = DictionaryList(0)
    distances = DictionaryList([])
    resNames = {}

    for frameIndex, frame in enumerate(frames):
        frameCombinations = []

        for bridge in frame:
            proteinAtoms = bridge.proteins
            for atom_1, atom_2 in combinations(proteinAtoms, r=2):
                ind_1, ind_2 = atom_1.getIndex(), atom_2.getIndex()
                res_1, res_2 = atom_1.getResnum(), atom_2.getResnum()

                if res_1 == res_2:
                    continue

                if (res_1, res_2) not in frameCombinations:
                    interactionCount[(res_1, res_2)] += 1
                    interactionCount[(res_2, res_1)] += 1
                    frameCombinations += [(res_1, res_2), (res_2, res_1)]

                coords = allCoordinates[frameIndex]
                atom_1_coords, atom_2_coords = coords[ind_1], coords[ind_2]
                distance = calcDistance(atom_1_coords, atom_2_coords)
                distances[(res_1, res_2)] += [distance]
                distances[(res_2, res_1)] += [distance]

                res_1_name, res_2_name = getResidueName(
                    atom_1),  getResidueName(atom_2)
                resNames[res_1] = res_1_name
                resNames[res_2] = res_2_name

    interactionCount.removeDuplicateKeys(
        lambda keys, key: (key[1], key[0]) in keys)

    tableHeader = '{0:<15}{1:<15}{2:<10}{3:<10}{4:<10}'.format(
        "RES1", "RES2", "PERC", "DIST_AVG", "DIST_STD")
    LOGGER.info(tableHeader)
    info = {}
    file = open(filename, 'w') if filename else None
    if file:
        file.write(tableHeader + '\n')

    for key in interactionCount.keys():
        percentage = 100 * interactionCount[key]/len(frames)
        distAvg = np.average(distances[key])
        distStd = np.std(distances[key])
        pairInfo = {"percentage": percentage,
                    "distAvg": distAvg, "distStd": distStd}

        outputKey = key
        x, y = key
        if output == 'info':
            outputKey = (resNames[x], resNames[y])
            info[outputKey] = pairInfo
        elif output == 'indices':
            key1, key2 = (x, y), (y, x)
            info[key1], info[key2] = pairInfo, pairInfo

        tableRow = '{0:<15}{1:<15}{2:<10.3f}{3:<10.3f}{4:<10.3f}'.format(
            resNames[x], resNames[y], percentage, distAvg, distStd)
        LOGGER.info(tableRow)
        if file:
            file.write(tableRow + '\n')

    if file:
        file.close()

    return info


def calcWaterBridgeMatrix(data, metric):
    """Returns matrix which has metric as value and residue ids as ax indices.

    :arg data: dictionary returned by calcWaterBridgesStatistics, output='indices' 
    :type data: dict

    :arg metric: dict key from data
    :type metric: 'percentage' | 'distAvg' | 'distStd'
    """
    maxResnum = max(max(key) for key in data.keys()) + 1
    resMatrix = np.zeros((maxResnum, maxResnum), dtype=float)

    for key, value in data.items():
        resMatrix[key] = value[metric]

    return resMatrix


def showWaterBridgeMatrix(data, metric):
    """Shows matrix which has percentage/avg distance as value and residue ids as ax indices.

    :arg data: dictionary returned by calcWaterBridgesStatistics, output='indices' 
    :type data: dict

    :arg metric: dict key from data
    :type metric: 'percentage' | 'distAvg' | 'distStd'
    """
    import matplotlib.pyplot as plt
    matrix = calcWaterBridgeMatrix(data, metric)
    titles = {
        'percentage': 'Interaction percentage',
        'distAvg': 'Average distance',
        'distStd': 'Distance standard deviation'
    }

    showMatrix(matrix)
    plt.title(titles[metric])


def reduceTo1D(list, elementSel=lambda x: x, sublistSel=lambda x: x):
    return [elementSel(element) for sublist in list for element in sublistSel(sublist)]


def mofifyBeta(bridgeFrames, atoms):
    residueOccurances = {}

    for frame in bridgeFrames:
        frameResnums = set(reduceTo1D(
            frame, lambda a: a.getResnum(), lambda b: b.proteins))
        for res in frameResnums:
            residueOccurances[res] = residueOccurances.get(res, 0) + 1

    atoms.setBetas(0)
    for resnum, value in residueOccurances.items():
        residueAtoms = atoms.select(
            'resnum {0}'.format(resnum))
        beta = value/len(bridgeFrames)

        residueAtoms.setBetas(beta)


def calcBridgingResiduesHistogram(frames, **kwargs):
    """Calculates, plots and returns number of frames that each residue is involved in making water bridges, sorted by value. 

    :arg frames: list of water bridges from calcWaterBridgesTrajectory(), output='atomic'
    :type frames: list

    :arg clip: maximal number of residues on graph; to represent all set None
        default is 20
    :type clip: int

    :arg use_segname: whether to use segname to label residues
        default is False, because then the labels get long
    :type use_segname: bool
    """

    show_plot = kwargs.pop('show_plot', False)
    use_segname = kwargs.get('use_segname', False)

    clip = kwargs.pop('clip', 20)
    if clip == None:
        clip = 0

    residuesWithCount = {}
    for frame in frames:
        frameResidues = set(reduceTo1D(
            frame, lambda x: getResidueName(x, use_segname=use_segname),
            lambda wb: wb.proteins))

        for res in frameResidues:
            residuesWithCount[res] = residuesWithCount.get(res, 0) + 1

    sortedResidues = sorted(residuesWithCount.items(),
                            key=lambda r: r[1])

    labels, values = zip(*sortedResidues[-clip:])

    if show_plot:
        import matplotlib.pyplot as plt
        plt.figure(figsize=(5, 3 + 0.11 * len(labels)))
        plt.barh(labels, values)
        plt.xlabel('Number of frame appearances')
        plt.ylabel('Residue')
        plt.title('Water bridging residues')
        plt.tight_layout()
        plt.margins(y=0.01)
        plt.gca().xaxis.set_label_position('top')
        plt.gca().xaxis.tick_top()
        if SETTINGS['auto_show']:
            showFigure()

    return sortedResidues

def showBridgingResiduesHistogram(frames, **kwargs):
    kwargs['show_plot'] = True
    return calcBridgingResiduesHistogram(frames, **kwargs)


def getBridgingResidues(frames, residue):
    residuesWithCount = {}

    for frame in frames:
        frameResidues = []
        for bridge in frame:
            resNames = list(map(getResidueName, bridge.proteins))
            if residue in resNames:
                frameResidues += resNames

        if not frameResidues:
            continue

        frameResidues = set(frameResidues)
        frameResidues.remove(residue)
        for res in frameResidues:
            residuesWithCount[res] = residuesWithCount.get(res, 0) + 1

    sortedResidues = sorted(residuesWithCount.items(),
                            key=lambda r: r[1], reverse=True)
    return sortedResidues


def getWaterCountDistribution(frames, res_a, res_b):
    waters = []

    for frame in frames:
        for bridge in frame:
            resNames = list(map(getResidueName, bridge.proteins))
            if not (res_a in resNames and res_b in resNames):
                continue

            waterInvolvedCount = len(bridge.waters)
            waters.append(waterInvolvedCount)

    return waters


def getDistanceDistribution(frames, res_a, res_b, trajectory):
    distances = []
    allCoordinates = trajectory.getCoordsets()

    for frameIndex, frame in enumerate(frames):
        for bridge in frame:
            resNames = list(map(getResidueName, bridge.proteins))
            if not (res_a in resNames and res_b in resNames):
                continue

            for atom_1, atom_2 in combinations(bridge.proteins, r=2):
                coords = allCoordinates[frameIndex]
                atom_1_coords, atom_2_coords = coords[atom_1.getIndex(
                )], coords[atom_2.getIndex()]

                dist = calcDistance(atom_1_coords, atom_2_coords)
                distances.append(dist)

    return distances


def getResidueLocationDistribution(frames, res_a, res_b):
    locationInfo = {"backbone": 0, "side": 0}
    result = {res_a: locationInfo.copy(), res_b: locationInfo.copy()}

    for frame in frames:
        for bridge in frame:
            atoms_a = filter(lambda a: getResidueName(a)
                             == res_a, bridge.proteins)
            atoms_b = filter(lambda a: getResidueName(a)
                             == res_b, bridge.proteins)

            if not atoms_a or not atoms_b:
                continue

            def atomType(a): return 'backbone' if a.getName() in [
                'CA', 'C', 'N', 'O'] else 'side'
            for atom in atoms_a:
                result[res_a][atomType(atom)] += 1
            for atom in atoms_b:
                result[res_b][atomType(atom)] += 1

    return result


def calcWaterBridgesDistribution(frames, res_a, res_b=None, **kwargs):
    """Returns distribution for certain metric and plots if possible.

    :arg res_a: name of first residue
    :type frames: str

    :arg res_b: name of second residue
        default is None
    :type frames: str

    :arg metric: 'residues' returns names and frame count of residues interacting with res_a,
                'waters' returns water count for each bridge between res_a and res_b
                'distance' returns distance between each pair of protein atoms involved in bridge between res_a and res_b
                'location' returns dictionary with backbone/sidechain count information
    :type metric: 'residues' | 'waters' | 'distance' | 'location'

    :trajectory: DCD file - necessary for distance distribution

    :arg output: return 2D matrices or dictionary where key is residue info
        default is 'dict'
    :type output: 'dict' | 'indices'
    """
    show_plot = kwargs.pop('show_plot', False)
    metric = kwargs.pop('metric', 'residues')
    trajectory = kwargs.pop('trajectory', None)

    if metric == 'distance' and not trajectory:
        raise TypeError(
            'Distance distribution measurement needs trajectory argument!')

    methods = {
        'residues': lambda: getBridgingResidues(frames, res_a),
        'waters': lambda: getWaterCountDistribution(frames, res_a, res_b),
        'distance': lambda: getDistanceDistribution(frames, res_a, res_b, trajectory),
        'location': lambda: getResidueLocationDistribution(frames, res_a, res_b)
    }

    result = methods[metric]()

    if metric in ['waters', 'distance'] and show_plot:
        import matplotlib.pyplot as plt
        plt.hist(result, rwidth=0.95, density=True)
        plt.xlabel('Value')
        plt.ylabel('Probability')
        plt.title('Distribution: {0}'.format(metric))
        if SETTINGS['auto_show']:
            showFigure()

    return result

def showWaterBridgesDistribution(frames, res_a, res_b=None, **kwargs):
    kwargs['show_plot'] = True
    return calcWaterBridgesDistribution(frames, res_a, res_b, **kwargs)


def savePDBWaterBridges(bridges, atoms, filename):
    """Saves single PDB with occupancy on protein atoms and waters involved bridges.

    :arg bridges: atomic output from calcWaterBridges
    :type bridges: list

    :arg atoms: Atomic object from which atoms are considered
    :type atoms: :class:`.Atomic`

    :arg filename: name of file to be saved
    :type filename: string
    """
    atoms = atoms.copy()

    mofifyBeta([bridges], atoms)
    atoms.setOccupancies(0)
    atoms.select('beta = 1').setOccupancies(1)

    proteinAtoms = atoms.select('protein')
    waterOxygens = reduceTo1D(
        bridges, lambda w: w.getIndex(), lambda b: b.waters)
    waterAtoms = atoms.select(
        'same residue as water within 1.6 of index {0}'.format(
            " ".join(map(str, waterOxygens))))

    atomsToSave = proteinAtoms.toAtomGroup() + waterAtoms.toAtomGroup()
    return writePDB(filename, atomsToSave)


def savePDBWaterBridgesTrajectory(bridgeFrames, atoms, filename, trajectory=None, max_proc=1):
    """Saves one PDB per frame with occupancy and beta on protein atoms and waters forming bridges in frame.

    :arg bridgeFrames: atomic output from calcWaterBridgesTrajectory
    :type bridgeFrames: list

    :arg atoms: Atomic object from which atoms are considered
    :type atoms: :class:`.Atomic`

    :arg filename: name of file to be saved; must end in .pdb
    :type filename: string

    :arg trajectory: trajectory data (not needed for multi-model PDB)
    :type trajectory: :class:`.Trajectory', :class:`.Ensemble`, :class:`.Atomic`
    """
    if not trajectory and atoms.numCoordsets() < len(bridgeFrames):
        raise TypeError('Provide parsed trajectory!')

    filename = filename[:filename.rfind('.')]

    atoms = atoms.copy()
    mofifyBeta(bridgeFrames, atoms)

    def saveBridgesFrame(trajectory, atoms, frameIndex, frame):
        LOGGER.info('Frame: {0}'.format(frameIndex))
        if trajectory:
            coords = trajectory[frameIndex].getCoords()
            atoms.setCoords(coords)
        else:
            atoms.setACSIndex(frameIndex)

        waterAtoms = reduceTo1D(frame, sublistSel=lambda b: b.waters)
        waterResidues = atoms.select(
            'same residue as water within 1.6 of index {0}'.format(
                " ".join(map(lambda a: str(a.getIndex()), waterAtoms))))

        bridgeProteinAtoms = reduceTo1D(
            frame, lambda p: p.getResnum(), lambda b: b.proteins)
        atoms.setOccupancies(0)
        atoms.select('resid {0}'.format(
            " ".join(map(str, bridgeProteinAtoms)))).setOccupancies(1)

        atomsToSave = atoms.select(
            'protein').toAtomGroup() + waterResidues.toAtomGroup()

        if trajectory:
            writePDB('{0}_{1}.pdb'.format(filename, frameIndex),
                     atomsToSave)
        else:
            writePDB('{0}_{1}.pdb'.format(filename, frameIndex),
                     atomsToSave, csets=frameIndex)

    if max_proc == 1:
        for frameIndex, frame in enumerate(bridgeFrames):
            saveBridgesFrame(trajectory, atoms, frameIndex, frame)
    else:
        frameIndex = 0
        numFrames = len(bridgeFrames)
        while frameIndex < numFrames:
            processes = []
            for _ in range(max_proc):
                p = mp.Process(target=saveBridgesFrame, args=(trajectory, atoms, frameIndex,
                                                              bridgeFrames[frameIndex]))
                p.start()
                processes.append(p)

                frameIndex += 1
                if frameIndex >= numFrames:
                    break

            for p in processes:
                p.join()

def getBridgeIndicesString(bridge):
    return ' '.join(map(lambda a: str(a.getIndex()), bridge.proteins))\
        + '|'\
        + ' '.join(map(lambda a: str(a.getIndex()), bridge.waters))


def saveWaterBridges(atomicBridges, filename):
    """Save water bridges as information (.txt) or WaterBridges (.wb) parsable file.

    :arg atomicBridges: atomic output from calcWaterBridges/Trajectory
    :type atomicBridges: list

    :arg filename: path where file should be saved
    :type filename: string
    """
    isSingleBridge = isinstance(atomicBridges[0], AtomicOutput)
    isInfoOutput = filename.split('.')[-1] == 'txt'
    file = open(filename, 'w')

    if isSingleBridge:
        atomicBridges = [atomicBridges]

    if isInfoOutput:
        info = getWaterBridgesInfoOutput(atomicBridges)
        for frameIndex, frame in enumerate(info):
            file.write('FRAME {0}\n'.format(frameIndex))
            for bridge in frame:
                file.write(' '.join(map(str, bridge)) + '\n')

        return file.close()

    for frame in atomicBridges:
        for bridge in frame:
            line = getBridgeIndicesString(bridge)
            file.write(line + '\n')
        file.write('ENDFRAME\n')

    return file.close()


def getBridgeFromIndices(proteinIndices, waterIndices, atoms):
    proteins = list(map(lambda i: atoms[i], proteinIndices))
    waters = list(map(lambda i: atoms[i], waterIndices))
    atomicBridge = AtomicOutput(proteins, waters)

    return atomicBridge


def parseWaterBridges(filename, atoms):
    """Parse water bridges from .wb file saved by saveWaterBridges, returns atomic type.

    :arg filename: path of file where bridges are stored
    :type filename: string

    :arg atoms: Atomic object on which calcWaterBridges was performed
    :type atoms: :class:`.Atomic`
    """
    bridgesFrames = []
    currentBridges = []

    with open(filename, 'r') as file:
        for line in file:
            line = line.strip()
            if line == 'ENDFRAME':
                bridgesFrames.append(currentBridges)
                currentBridges = []
                continue

            proteinString, waterString = line.split('|')
            proteinIndices = map(int, proteinString.split())
            waterIndices = map(int, waterString.split())

            bridge = getBridgeFromIndices(proteinIndices, waterIndices, atoms)
            currentBridges.append(bridge)

    if len(bridgesFrames) == 1:
        bridgesFrames = bridgesFrames[0]

    return bridgesFrames


def findClusterCenters(file_pattern, **kwargs):
    """ Find molecules that are forming cluster in 3D space.
    
    :arg file_pattern: file pattern for anlaysis
        it can include '*'
        example:'file_*.pdb' will analyze file_1.pdb, file_2.pdb, etc.
    :type file_pattern: str
    
    :arg selection: selection string
        by default 'water and name "O.*"' is used
    :type selection: str
    
    :arg distC: distance to other molecules
    :type distC: int, float
        default is 0.3
        
    :arg numC: min number of molecules in a cluster
        default is 3
    :type numC: int

    :arg filename: filename for output pdb file with clusters
        Default of **None** leads to
        'clusters_'+file_pattern.split("*")[0]+'.pdb'
    :type filename: str
    """
    
    import glob
    import numpy as np

    selection = kwargs.pop('selection', 'water and name "O.*"')
    distC = kwargs.pop('distC', 0.3)
    numC = kwargs.pop('numC', 3)
    filename = kwargs.pop('filename', None)
    
    matching_files = glob.glob(file_pattern)
    matching_files.sort()
    coords_all = parsePDB(matching_files[0]).select(selection).toAtomGroup()

    for i in matching_files[1:]:
        coords = parsePDB(i).select(selection).toAtomGroup()
        coords_all += coords

    removeResid = []
    removeCoords = []
    for ii in range(len(coords_all)):
        if 'water' in selection.split() or np.any([water in selection.split() for water in DEFAULTS['water']]):
            sel = coords_all.select('water within '+str(distC)+' of center', 
                                    center=coords_all.getCoords()[ii])
        else:
            sel = coords_all.select(str(selection)+' within '+str(distC)+' of center',
                                    center=coords_all.getCoords()[ii])
        
        if sel is not None and len(sel) <= int(numC):
            removeResid.append(coords_all.getResnums()[ii])
            removeCoords.append(list(coords_all.getCoords()[ii]))

    if len(removeCoords) == coords_all.numAtoms():
        raise ValueError('No waters were selected. You may need to align your trajectory \
        or change default parameters for detecting water clusters (increase distC \
        or decrease numC)')

    selectedWaters = AtomGroup()
    sel_waters = [] 

    for j in coords_all.getCoordsets()[0].tolist():
        if j not in removeCoords:
            sel_waters.append(j)

    coords_wat = np.array([sel_waters], dtype=float)
    selectedWaters.setCoords(coords_wat)
    selectedWaters.setNames(['DUM']*len(selectedWaters))
    selectedWaters.setResnums(range(1, len(selectedWaters)+1))
    selectedWaters.setResnames(['DUM']*len(selectedWaters))

    if filename is None:
        try:
            filename = 'clusters_'+file_pattern.split("*")[0]+'.pdb'
        except:
            filename = 'clusters.pdb'

    writePDB(filename, selectedWaters)
    LOGGER.info("Results are saved in {0}.".format(filename))

def filterStructuresWithoutWater(structures, min_water=0, filenames=None):
    """This function will filter out structures from *structures* that have no water 
    or fewer water molecules than *min_water*.
    
    :arg structures: list of :class:`.Atomic` structures to be filtered
    :type structures: list

    :arg min_water: minimum number of water O atoms, 
        default is 0
    :type min_water: int

    :arg filenames: an optional list of filenames to filter too
        This is an output argument
    :type filenames: list
    """

    if not isinstance(structures, list):
        raise TypeError('structures should be a list')
    
    if not np.all([isinstance(struct, Atomic) for struct in structures]):
        raise ValueError('elements of structures should be Atomic objects')
    
    if not isinstance(min_water, int):
        raise TypeError('min_water should be an integer')
    
    if filenames is None: filenames = []

    if not isinstance(filenames, list):
        raise TypeError('filenames should be None or a list')
    
    if len(filenames) not in [0, len(structures)]:
        raise TypeError('filenames should have the same length as structures')
    
    if not np.all([isinstance(filename, str) for filename in filenames]):
        raise ValueError('elements of filenames should be strings')
    
    if not np.all([os.path.exists(filename) for filename in filenames]):
        raise ValueError('at least one of the filenames does not exist')
    
    have_filenames = len(filenames)>0

    new_structures = []
    numStructures = len(structures)
    for i, struct in enumerate(reversed(structures)):
        title = struct.getTitle()
        waters = struct.select('water and name O')
    
        if waters == None:
            LOGGER.warn(title+" doesn't contain water molecules")
            if have_filenames:
                filenames.pop(numStructures-i-1)
            continue
    
        numWaters = waters.numAtoms()
        if numWaters < min_water:
            LOGGER.warn(title+" doesn't contain enough water molecules ({0})".format(numWaters))
            if have_filenames:
                filenames.pop(numStructures-i-1)
            continue

        new_structures.append(struct)

    return list(reversed(new_structures))


def selectSurroundingsBox(atoms, select, **kwargs):
    """Select the surroundings of *select* within *atoms* using
    a bounding box with optional *padding*.

    :arg return_selstr: whether to return the final selstr
        Default False
    :type return_selstr: bool
    """

    if not isinstance(atoms, Atomic):
        raise TypeError('atoms should be an Atomic object')

    if isinstance(select, str):
        select = atoms.select(select)

    if not isinstance(select, Atomic):
        raise TypeError('select should be a valid selection or selection string')

    padding = kwargs.get('padding', 0)
    if not isinstance(padding, Number):
        raise TypeError('padding should be a number')
    if padding < 0:
        raise ValueError('padding should be a positive number')

    return_selstr = kwargs.get('return_selstr', False)
    if not isinstance(return_selstr, bool):
        raise TypeError('return_selstr should be a bool')

    minCoords = select.getCoords().min(axis=0)
    maxCoords = select.getCoords().max(axis=0)

    if padding > 0:
        minCoords -= padding
        maxCoords += padding

    selstr = 'same residue as ((x `{0} to {1}`) and (y `{2} to {3}`) and (z `{4} to {5}`))'.format(
        minCoords[0], maxCoords[0], minCoords[1], maxCoords[1], minCoords[2], maxCoords[2])

    if return_selstr:
        return selstr
    return atoms.select(selstr)


def findCommonSelectionTraj(atoms, traj, selstr, **kwargs):
    """Select *selstr* within *atoms* for each frame in *traj*
    using a bounding box with optional *padding*.

    :arg expand_selection: whether to expand selections with
        :meth:`.selectSurroundingsBox`. Default False
    :type expand_selection: bool

    Returns the common selection and corresponding indices and
    optionally the corresponding selstr if *return_selstr* is **True**
    """

    if not isinstance(atoms, Atomic):
        raise TypeError('atoms should be an Atomic object')

    if not isinstance(selstr, str):
        raise TypeError('selstr should be a string')

    expand_selection = kwargs.get('expand_selection', False)
    if not isinstance(expand_selection, bool):
        raise TypeError('expand_selection should be a bool')

    return_selstr = kwargs.pop('return_selstr', False) # not passed on
    if not isinstance(return_selstr, bool):
        raise TypeError('return_selstr should be a bool')

    indices = None
    if selstr is not None:
        LOGGER.info('Finding common selection')
        indices = []
        for frame0 in traj:
            atoms_copy = atoms.copy()
            atoms_copy.setCoords(frame0.getCoords())
            selection = atoms_copy.select(selstr)

            if expand_selection:
                selection = selectSurroundingsBox(atoms_copy, selection, **kwargs)

            indices.extend(list(selection.getIndices()))

        indices = np.unique(indices)
        selection = atoms_copy[indices]

        if return_selstr:
            return selection, indices, selstr
        return selection, indices



# -*- coding: utf-8 -*-
"""This module defines miscellaneous functions dealing with protein data."""

import numpy as np

from prody.atomic import Atomic, Atom, AtomGroup, Selection, HierView
from prody.utilities import openFile, showFigure, createStringIO, wrapModes
from prody import SETTINGS, PY3K

__all__ = ['view3D', 'showProtein']

def view3D(*alist, **kwargs):
    """Return a py3Dmol view instance for interactive visualization in
    Jupyter notebooks. Available arguments are: width, height (of
    the viewer), backgroundColor, zoomTo (a py3Dmol selection to center
    around), and style, which is a py3Dmol style object that will be
    applied to all atoms in the scene. More complex styling can be achieved
    by manipulating the view object directly.
    
    The default style is to show the protein in a rainbow cartoon and
    hetero atoms in sticks/spheres.
    
    GNM/ANM Coloring
    
    An array of fluctuation values can be provided with the *data* kwarg
    for visualization of GNM/ANM calculations.  The array is assumed to 
    correpond to a calpha selection of the provided protein.
    The default color will be set to a RWB color scheme on a per-residue
    basis.  If the fluctuation vector contains negative values, the
    midpoint (white) will be at zero.  Otherwise the midpoint is the mean.
    
    An array of displacement vectors can be provided with the *mode* kwarg.
    The animation of these motions can be controlled with frames (number
    of frames to animate over), amplitude (scaling factor), and animate
    (3Dmol.js animate options).  If animation isn't enabled, by default
    arrows are drawn.  Drawing of arrows is controlled by the boolean arrows
    option and the arrowcolor option.
    
    If multiple structures are provided with the data or mode arguments, these
    arguments must be provided as lists of arrays of the appropriate dimension.

    If a 3Dmol.js viewer as specified as the view argument, that viewer will be
    modified and returned.  After modification, update instead of show should
    be called on the viewer object if it is desired to update in-place
    instead of instantiating a new viewer.
    """

    try:
        import py3Dmol
    except:
        raise ImportError('py3Dmol needs to be installed to use view3D')
    
    from .pdbfile import writePDBStream
    
    width = kwargs.get('width', 400)
    height = kwargs.get('height', 400)
    data_list = kwargs.pop('data', None)
    modes = kwargs.pop('mode', None)
    style = kwargs.pop('style', [])
    zoomto = kwargs.pop('zoomto', {})
    bgcolor = kwargs.pop('backgroundcolor', 'white')
    bgcolor = kwargs.pop('backgroundColor', bgcolor)
    frames = kwargs.pop('frames', 30)
    interval = kwargs.pop('interval', 1)
    anim = kwargs.pop('anim', False)
    scale = kwargs.pop('scale', 100)
    arrows = kwargs.pop('arrows',True)
    arrowcolor = kwargs.pop('arrowcolor', 'darkgrey')
    arrowcolor = kwargs.pop('arrowColor', arrowcolor)

    if modes is None:
        n_modes = 0
    else:
        modes = wrapModes(modes)
        n_modes = len(modes)

    if data_list is None:
        n_data = 0
    else:
        data_list = wrapModes(data_list)
        n_data = len(data_list)

    view = kwargs.get('view', None)
    js = kwargs.get('js', 'https://3dmol.csb.pitt.edu/build/3Dmol-min.js')

    if view is None:
        view = py3Dmol.view(width=width, height=height, js=js)

    def _mapData(atoms, data):
        # construct map from residue to data property
        propmap = []
        for j, a in enumerate(atoms.calpha):
            propmap.append({'model': -1, 'chain': a.getChid(), 'resi': int(a.getResnum()), 
                            'props': {'data': float(data[j]) } })
        # set the atom property 
        # TODO: implement something more efficient on the 3Dmol.js side (this is O(n*m)!)
        view.mapAtomProperties(propmap)
        
        # color by property using gradient
        extreme = np.max(np.abs(data))
        lo = -extreme if np.min(data) < 0 else 0
        mid = np.mean(data) if np.min(data) >= 0 else 0
        view.setColorByProperty({'model': -1}, 'data', 'rwb', [extreme,lo,mid])
        view.setStyle({'model': -1},{'cartoon':{'style':'oval'}})    


    for i, atoms in enumerate(alist):
        pdb = createStringIO()
        writePDBStream(pdb, atoms)
        view.addAsOneMolecule(pdb.getvalue(), 'pdb')
        view.setStyle({'model': -1}, {'cartoon': {'color':'spectrum'}})
        view.setStyle({'model': -1, 'hetflag': True}, {'stick':{}})
        view.setStyle({'model': -1, 'bonds': 0}, {'sphere':{'radius': 0.5}})    

        if n_data:
            data = data_list[i]
            try:
                data = data.getArray()
            except AttributeError:
                pass

            if atoms.calpha.numAtoms() != len(data):
                raise RuntimeError("Atom count mismatch: {} vs {}. data styling assumes a calpha selection."
                                   .format(atoms.calpha.numAtoms(), len(data)))
            _mapData(atoms, data)
            
        if n_modes:
            mode = modes[i]
            try:
                arr = mode.getArray()
                is3d = mode.is3d()
            except AttributeError:
                arr = mode
                is3d = len(arr) == atoms.calpha.numAtoms()*3

            if is3d:
                if atoms.calpha.numAtoms()*3 != len(arr):
                    raise RuntimeError("Atom count mismatch: {} vs {}. mode animation assumes a calpha selection."
                                       .format(atoms.calpha.numAtoms(), len(arr)//3))

                if anim:
                    # construct map from residue to anm property and dx,dy,dz vectors
                    propmap = []
                    for j, a in enumerate(atoms.calpha):
                        propmap.append({'model': -1, 'chain': a.getChid(), 'resi': int(a.getResnum()),
                                        'props': {'dx': arr[3*j], 'dy': arr[3*j+1], 'dz': arr[3*j+2] } })
                    # set the atom property 
                    # TODO: implement something more efficient on the 3Dmol.js side (this is O(n*m)!)
                    view.mapAtomProperties(propmap)
                elif arrows:
                    for j, a in enumerate(atoms.calpha):
                        start = a._getCoords()
                        dcoords = arr[3*j:3*j+3]
                        end = start + dcoords * scale
                        view.addArrow({'start': {'x':start[0], 'y':start[1], 'z':start[2]},
                                       'end': {'x':end[0], 'y':end[1], 'z':end[2]},
                                       'radius': 0.3, 'color': arrowcolor})
            else:
                if atoms.calpha.numAtoms() != len(arr):
                    raise RuntimeError("Atom count mismatch: {} vs {}. mode styling assumes a calpha selection."
                                       .format(atoms.calpha.numAtoms(), len(arr)))

                _mapData(atoms, arr)

                
    # setting styles ...
    view.setBackgroundColor(bgcolor)

    if n_modes and anim:
        #create vibrations
        view.vibrate(frames, scale)
        
        animate = kwargs.get('animate', {'loop':'rock', 'interval':interval})
        view.animate(animate)     

    if isinstance(style, dict):
        style = ({}, style)
    if isinstance(style, tuple):
        styles = [style]
    else:
        styles = style
    for sel, style in styles:
        view.setStyle(sel, style)
        
    view.zoomTo(zoomto)
            
    return view

def showProtein(*atoms, **kwargs):
    """Show protein representation using :meth:`~mpl_toolkits.mplot3d.Axes3D`.
    This function is designed for generating a quick view of the contents of a
    :class:`~.AtomGroup` or :class:`~.Selection`.    

    Protein atoms matching ``"calpha"`` selection are displayed using solid
    lines by picking a random and unique color per chain.  Line with can
    be adjusted using *lw* argument, e.g. ``lw=12``. Default width is 4.
    Chain colors can be overwritten using chain identifier as in ``A='green'``.

    Water molecule oxygen atoms are represented by red colored circles.  Color
    can be changed using *water* keyword argument, e.g. ``water='aqua'``.
    Water marker and size can be changed using *wmarker* and *wsize* keywords,
    defaults values are ``wmarker='.', wsize=6``.

    Hetero atoms matching ``"hetero and noh"`` selection are represented by
    circles and unique colors are picked at random on a per residue basis.
    Colors can be customized using residue name as in ``NAH='purple'``.  Note
    that this will color all distinct residues with the same name in the same
    color.  Hetero atom marker and size can be changed using *hmarker* and
    *hsize* keywords, default values are ``hmarker='o', hsize=6``.

    ProDy will set the size of axis so the representation is not distorted when
    the shape of figure window is close to a square.  Colors are picked at
    random, except for water oxygens which will always be colored red.
    
    *** Interactive 3D Rendering in Jupyter Notebook ***
    
    If py3Dmol has been imported then it will be used instead to display 
    an interactive viewer.  See :func:`view3D`
    
    
    """

    from prody.dynamics.mode import Mode

    method = kwargs.pop('draw', None)
    modes = kwargs.get('mode', None)
    scale = kwargs.get('scale', 100)

    # modes need to be specifically a list or a tuple (cannot be an array)
    if modes is None:
        n_modes = 0
    else:
        modes = wrapModes(modes)
        n_modes = len(modes)

    if method is None:
        import sys        
        if 'py3Dmol' in sys.modules: 
            method = 'py3Dmol'
        else:
            method = 'matplotlib'
    method = method.lower()
        
    alist = atoms
    for atoms in alist:
        if not isinstance(atoms, Atomic):
            raise TypeError('atoms must be an Atomic instance')
            
    if n_modes and n_modes != len(alist):
        raise RuntimeError('the number of proteins ({0}) does not match that of the modes ({1}).'
                            .format(len(alist), n_modes))

    if '3dmol' in method:
        mol = view3D(*alist, **kwargs)
        return mol
    else:
        kwargs.pop('mode', None)
        kwargs.pop('scale', 100)

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        cf = plt.gcf()
        show = None
        for child in cf.get_children():
            if isinstance(child, Axes3D):
                show = child
                break
        if show is None:
            show = cf.add_subplot(1,1,1,projection="3d")
        from matplotlib import colors
        cnames = dict(colors.cnames)
        wcolor = kwargs.get('water', 'red').lower()
        avoid = np.array(colors.hex2color(cnames.pop(wcolor, cnames.pop('red'))))
        for cn, val in cnames.copy().items():  # PY3K: OK
            clr = np.array(colors.hex2color(val))
            if clr.sum() > 2.4:
                cnames.pop(cn)
            elif np.abs(avoid - clr).sum() <= 0.6:
                cnames.pop(cn)
        cnames = list(cnames)
        import random
        random.shuffle(cnames)
        cnames_copy = list(cnames)
        min_ = list()
        max_ = list()
        for i, atoms in enumerate(alist):
            if isinstance(atoms, AtomGroup):
                title = atoms.getTitle()
            else:
                title = atoms.getAtomGroup().getTitle()
            calpha = atoms.select('calpha')
            if calpha:
                partition = False
                mode = modes[i] if n_modes else None
                if mode is not None:
                    is3d = False
                    try:
                        arr = mode.getArray()
                        is3d = mode.is3d()
                        n_nodes = mode.numAtoms()
                    except AttributeError:
                        arr = mode
                        is3d = len(arr) == len(calpha)*3
                        n_nodes = len(arr)//3 if is3d else len(arr)
                    if n_nodes != len(calpha):
                        raise RuntimeError('size mismatch between the protein ({0} residues) and the mode ({1} nodes).'
                                            .format(len(calpha), n_nodes))
                    partition = not is3d

                if partition:
                    xyz = calpha._getCoords()
                    chids = calpha.getChids()
                    rbody = []
                    last_sign = np.sign(arr[0])
                    last_chid = chids[0]
                    rcolor = ['red', 'red', 'blue']
                    n = 1
                    for i,a in enumerate(arr):
                        s = np.sign(a)
                        ch = chids[i]
                        if s == 0: s = last_sign
                        if last_sign != s or i == len(arr)-1 or last_chid != ch:
                            if last_chid == ch:
                                rbody.append(i)
                            show.plot(xyz[rbody, 0], xyz[rbody, 1], xyz[rbody, 2],
                                      label=title + '_regid%d'%n,
                                      color=rcolor[int(last_sign+1)],
                                      lw=kwargs.get('lw', 4))
                            rbody = []
                            n += 1
                            last_sign = s
                            last_chid = ch
                        rbody.append(i)
                else:
                    for ch in HierView(calpha, chain=True):
                        xyz = ch._getCoords()
                        chid = ch.getChid()
                        if len(cnames) == 0:
                            cnames = list(cnames_copy)
                        show.plot(xyz[:, 0], xyz[:, 1], xyz[:, 2],
                                label=title + '_' + chid,
                                color=kwargs.get(chid, cnames.pop()).lower(),
                                lw=kwargs.get('lw', 4))
                    
                    if mode is not None:
                        from prody.utilities.drawtools import drawArrow3D
                        XYZ = calpha._getCoords()
                        arr = arr.reshape((n_nodes, 3))
                        XYZ2 = XYZ + arr * scale
                        for i, xyz in enumerate(XYZ):
                            xyz2 = XYZ2[i]
                            mutation_scale = kwargs.pop('mutation_scale', 10)
                            drawArrow3D(xyz, xyz2, mutation_scale=mutation_scale, **kwargs)

            water = atoms.select('water and noh')
            if water:
                xyz = atoms.select('water')._getCoords()
                show.plot(xyz[:, 0], xyz[:, 1], xyz[:, 2], label=title + '_water',
                          color=wcolor,
                          ls='None', marker=kwargs.get('wmarker', '.'),
                          ms=kwargs.get('wsize', 6))
            hetero = atoms.select('not protein and not nucleic and not water and not dummy')
            if hetero:
                for res in HierView(hetero).iterResidues():
                    xyz = res._getCoords()
                    resname = res.getResname()
                    resnum = str(res.getResnum())
                    chid = res.getChid()
                    if len(cnames) == 0:
                        cnames = list(cnames_copy)
                    show.plot(xyz[:, 0], xyz[:, 1], xyz[:, 2], ls='None',
                              color=kwargs.get(resname, cnames.pop()).lower(),
                              label=title + '_' + chid + '_' + resname + resnum,
                              marker=kwargs.get('hmarker', 'o'),
                              ms=kwargs.get('hsize', 6))
            xyz = atoms._getCoords()
            min_.append(xyz.min(0))
            max_.append(xyz.max(0))

        show.set_xlabel('x')
        show.set_ylabel('y')
        show.set_zlabel('z')
        min_ = np.array(min_).min(0)
        max_ = np.array(max_).max(0)
        center = (max_ + min_) / 2
        half = (max_ - min_).max() / 2
        show.set_xlim3d(center[0]-half, center[0]+half)
        show.set_ylim3d(center[1]-half, center[1]+half)
        show.set_zlim3d(center[2]-half, center[2]+half)
        if kwargs.get('legend', False):
            show.legend(prop={'size': 10})
        if SETTINGS['auto_show']:
            showFigure()
        return show



# -*- coding: utf-8 -*-
"""This module defines classes for handling polypeptide/nucleic acid chains."""

from numpy import arange, unique

from .subset import AtomSubset
from .atomic import AAMAP

__all__ = ['Chain']

def getSequence(resnames, **kwargs):
    """Returns polypeptide sequence from a list of *resnames* using one-letter residue
    name abbreviations by default, or long (usually three letter) abbrevations
    if *longSeq* is **True**."""

    longSeq = kwargs.get('longSeq', False)
    if longSeq:
        return ' '.join(resnames)

    get = AAMAP.get
    return ''.join([get(rn, 'X') for rn in resnames])


class Chain(AtomSubset):

    """Instances of this class point to atoms with same chain identifiers and
    are generated by :class:`.HierView` class.  Following built-in functions
    are customized for this class:

    * :func:`len` returns the number of residues in the chain
    * :func:`iter` yields :class:`.Residue` instances

    Indexing :class:`Chain` instances by:

         - *residue number [, insertion code]* (:func:`tuple`),
           e.g. ``10`` or  ``10, "B"``, returns a :class:`.Residue`
         - *slice* (:func:`slice`), e.g, ``10:20``, returns a list of
           :class:`.Residue` instances"""

    __slots__ = ['_ag', '_indices', '_hv', '_acsi', '_selstr', '_seq']

    def __init__(self, ag, indices, hv, acsi=None, **kwargs):

        AtomSubset.__init__(self, ag, indices, acsi, **kwargs)
        self._hv = hv
        self._seq = None

    def __repr__(self):

        n_csets = self._ag.numCoordsets()
        segment = self.getSegment()
        if segment is None:
            segment = ''
        else:
            segment = ' from ' + str(segment)
        if n_csets == 1:
            return ('<Chain: {0}{1} from {2} ({3} residues, {4} atoms)>'
                    ).format(self.getChid(), segment, self._ag.getTitle(),
                             self.numResidues(), self.numAtoms())
        elif n_csets > 1:
            return ('<Chain: {0}{1} from {2} ({3} residues, {4} '
                    'atoms; active #{5} of {6} coordsets)>'
                    ).format(self.getChid(), segment, self._ag.getTitle(),
                             self.numResidues(), self.numAtoms(),
                             self.getACSIndex(), n_csets)
        else:
            return ('<Chain: {0}{1} from {2} ({3} residues, '
                    '{4} atoms; no coordinates)>'
                    ).format(self.getChid(), segment, self._ag.getTitle(),
                             self.numResidues(), self.numAtoms())

    def __str__(self):

        return 'Chain ' + self.getChid()

    def __getitem__(self, key):

        if isinstance(key, tuple):
            return self.getResidue(*key)
        else:
            return AtomSubset.__getitem__(self, key)

    def getSegment(self):
        """Returns segment of the chain."""

        segname = self.getSegname()
        if segname is not None:
            return self._hv.getSegment(segname)

    def getSegname(self):
        """Returns segment name."""

        segnames = self._ag._getSegnames()
        if segnames is not None:
            return segnames[self._indices[0]]

    def getResidue(self, resnum, icode=None):
        """Returns residue with number *resnum* and insertion code *icode*."""

        return self._hv.getResidue(self.getChid(), resnum, icode,
                                   self.getSegname())

    def iterResidues(self):
        """Yield residues."""

        get = self._hv._getResidue
        for index in unique(self._getResindices()):
            yield get(index)

    __iter__ = iterResidues

    def numResidues(self):
        """Returns number of residues."""

        return len(set(self._getResindices()))

    __len__ = numResidues

    def getChid(self):
        """Returns chain identifier."""

        return self._ag._getChids()[self._indices[0]]

    def setChid(self, chid):
        """Set chain identifier."""

        self.setChids(chid)

    def getChindex(self):
        """Returns chain index."""

        return self._ag._getChindices()[self._indices[0]]

    def getSequence(self, **kwargs):
        """Returns one-letter sequence string for amino acids in the chain.
        When *allres* keyword argument is **True**, sequence will include all
        residues (e.g. water molecules) in the chain and **X** will be used for
        non-standard residue names."""

        if kwargs.get('allres', False):
            get = AAMAP.get
            if kwargs.get('longSeq', False):
                seq = ' '.join([res.getResname() for res in self])
            else:
                seq = ''.join([get(res.getResname(), 'X') for res in self])
        elif self._seq:
            seq = self._seq
        else:
            calpha = self.calpha
            if calpha:
                seq = getSequence(calpha.getResnames(), **kwargs)
            else:
                seq = ''
            self._seq = seq
        return seq

    def getSelstr(self):
        """Returns selection string that selects atoms in this chain."""

        segment = self.getSegment()
        if segment is None:
            if self._selstr:
                return 'chain {0} and ({1})'.format(self.getChid(),
                                                    self._selstr)
            else:
                return 'chain {0}'.format(self.getChid())
        else:
            return 'chain {0} and ({1})'.format(self.getChid(),
                                                segment.getSelstr())

    def getHierView(self, **kwargs):
        """Returns a hierarchical view of the this chain."""

        return HierView(self, **kwargs)
        